#include <string.h>
#include <algorithm>

#include <osg/GL>
#include <osg/PolygonOffset>
#include <osg/Material>
#include <osg/LineWidth>
#include <osg/PolygonMode>
#include <osgViewer/Viewer>
#include <osg/Node>
#include <osg/Node>
#include <osg/Group>
#include <osg/NodeVisitor>
#include <osg/Notify>
#include <osg/OccluderNode>
#include <osg/Transform>

#include "ColladaEffectNode.h"

#include <dae.h>

#include "dom/domGlsl_surface_type.h"

/// Helper class for generating NodePathList.
class MyCollectParentPaths : public osg::NodeVisitor
    {
    public:
        MyCollectParentPaths(osg::Node* haltTraversalAtNode=0) : 
            osg::NodeVisitor(osg::NodeVisitor::TRAVERSE_PARENTS),
            _haltTraversalAtNode(haltTraversalAtNode)
        {
        }

        virtual void apply(osg::Node& node)
        {
            if (node.getNumParents()==0 || &node==_haltTraversalAtNode)
            {
                _nodePaths.push_back(getNodePath());
            }
            else
            {
                traverse(node);
            }
       }

        osg::Node*           _haltTraversalAtNode;
        osg::NodePath        _nodePath;
        osg::NodePathList    _nodePaths;
    };

ColladaEffectNode::ColladaEffectNode()
:  m_fTimeCycle( fDefaultRenderMonkeyTimeCycle ),
   osg::Group(),
   m_pBaseGeometry(NULL)
{
   for ( unsigned int i = 0 ; i < 128 ; i++ )
   {
      m_PassIsRtt[i] = false;
   }
}

ColladaEffectNode::ColladaEffectNode( const ColladaEffectNode& copy, const osg::CopyOp& copyop )
: m_fTimeCycle( fDefaultRenderMonkeyTimeCycle )
{
}

ColladaEffectNode::~ColladaEffectNode()
{
   // disable the validator for safety, so it won't try to access us
   // even if it stays alive for some reason
}

ColladaRTT * ColladaEffectNode::AddRenderTarget (  const std::string    &  strSurfaceSID, 
                                                   int                     nPassIndex )
{
   ColladaRTT * pRTT = NULL ;

   RTTNameMap::iterator it = m_RTTNameMap.find( strSurfaceSID );

   // If not found then create a new one.
   if ( it == m_RTTNameMap.end() )
   {
      pRTT = new ColladaRTT( strSurfaceSID, nPassIndex ) ;
      if ( !pRTT ) 
      {
         osg::notify( osg::WARN ) << "AddRenderTarget: Could not create RTT object." << std::endl;
         return NULL;
      }
      m_RTTIndexMap[ nPassIndex ] = pRTT ;
      m_RTTNameMap[ strSurfaceSID ] = pRTT ;
      m_RTTVector.push_back ( pRTT );
      m_PassIsRtt[nPassIndex] = true ;

      //m_bIsRTT = true ;
   }
   else
   {
      pRTT = it->second;
   }

   return pRTT;
}

ColladaRTT * ColladaEffectNode::GetRenderTargetByName ( std::string strName )
{
   ColladaRTT * pRTT = NULL ;

   RTTNameMap::iterator it = m_RTTNameMap.find( strName );
   if ( it != m_RTTNameMap.end() )
   {
      pRTT = it->second ;
   }

   return pRTT;
}

ColladaRTT * ColladaEffectNode::GetRenderTargetByIndex ( unsigned int nIndex )
{
   ColladaRTT * pRTT = NULL ;

   RTTIndexMap::iterator it = m_RTTIndexMap.find( nIndex );
   if ( it != m_RTTIndexMap.end() )
   {
      pRTT = it->second ;
   }

   return pRTT;
}

void ColladaEffectNode::AddGeometry ( osg::Node * pGeom )
{
   unsigned int nCount = m_RTTVector.size();
   if ( nCount ) 
   {
      for ( unsigned int i = 0 ; i < nCount ; i++ )
      {
         ColladaRTT *pRTT = GetRenderTargetByIndex( i );
         if ( pRTT )
         {
            // Add the RTT root node
            // Add the geometry to the RTT node.
            pRTT->AddTextureGeometry( pGeom ) ;
         }
      }
   }

   // Add the geometry to the node (as well as to the RTT nodes)
   this->asGroup()->addChild( pGeom );
}

void ColladaEffectNode::PrintMatrix ( osg::Matrixf & mat )
{
   osg::Vec3f eye, center, up ;
   float fDistance = -1.0f ;
   mat.getLookAt ( eye, center, up, fDistance );

   osg::notify( osg::WARN ) << "Eye: " << eye[0] << " " << eye[1] << " " << eye[2] << " " << std::endl;
   osg::notify( osg::WARN ) << "At : " << center[0] << " " << center[1] << " " << center[2] << " " <<  std::endl;
   osg::notify( osg::WARN ) << "Up : " << up[0] << " " << up[1] << " " << up[2] << " " <<  std::endl;
}

void ColladaEffectNode::HandleUpdate ( osg::NodeVisitor& nv )
{
   osg::Group *pParent = this;

   osgUtil::UpdateVisitor *uv = dynamic_cast<osgUtil::UpdateVisitor *>(&nv);
   if ( !uv )
   {
      // We are only interested in updates traversals.
      return;
   }

   RTTVector::iterator it = m_RTTVector.begin() ;      
   while ( it != m_RTTVector.end() )
   {
      ColladaRTT * pRTT = *it;
      // Only affect RTTs that change their camera position.
      if ( pRTT && pRTT->GetMainViewLock() )
      {
         // If m_bMainViewLock is true in the ColladaRTT object it means we want to pass the
         // motion of the main camera down to the RTT camera. So, we need to find out what the 
         // view transformation currently is. 
         osg::Matrixf  masterCameraVPW = GetWorldMatrices ( nv.getNodePath().back(), false )[0];

         // Pass the update and main camera view to the RTT object.
         pRTT->HandleUpdate ( nv, masterCameraVPW );
      }
      ++it;
   }
} 

// This code is taken from osg::Node
osg::MatrixList ColladaEffectNode::GetWorldMatrices( osg::Node* haltTraversalAtNode, bool bIgnoreCameras ) const
{
   using namespace osg;

    MyCollectParentPaths cpp(NULL);
    const_cast<Node*>(haltTraversalAtNode)->accept(cpp);
    
    MatrixList matrices;
    
    for( NodePathList::iterator itr = cpp._nodePaths.begin();
        itr != cpp._nodePaths.end();
        ++itr )
    {
        NodePath& nodePath = *itr;
        if (nodePath.empty())
        {
            matrices.push_back(osg::Matrix::identity());
        }
        else
        {
            matrices.push_back(osg::computeLocalToWorld(nodePath, bIgnoreCameras));
        }
    }
    
    return matrices;
}

void ColladaEffectNode::traverse( osg::NodeVisitor& nv )
{
   // Pass the update on to derived classes so they can handle the update the way they like.
   HandleUpdate ( nv );

   //osgUtil::CullVisitor *cv = dynamic_cast<osgUtil::CullVisitor *>(&nv);
   //if ( cv )
   //{
   //printf ( "Effect::CullVisitor frame: %d\n", uv->getFrameStamp()->getFrameNumber() );
   //}
   //osgUtil::UpdateVisitor *uv = dynamic_cast<osgUtil::UpdateVisitor *>(&nv);
   //if ( uv )
   //{
   //   printf ( "Effect::UpdateVisitor frame: %d\n", uv->getFrameStamp()->getFrameNumber() );
   //}

   // if we could find an active technique, then continue with traversal,
   // else go for default traversal (no effect)
   if ( _techs[0].get() ) 
   {
      _techs[0]->traverse( nv, this );
   } 
   else 
   {
      if ( nv.getTraversalMode() == osg::NodeVisitor::TRAVERSE_ALL_CHILDREN ) 
      {
         inherited_traverse( nv );
      }
   }
}

void ColladaEffectNode::inherited_traverse( osg::NodeVisitor& nv, int nPassIndex )
{
   // If we are not an RTT just traverse the group's children.
   if ( false == m_PassIsRtt[nPassIndex] )
   {
      osg::Group::traverse(nv);
      return;
   }

   // Only render the correct RTT 
   if ( nPassIndex > -1 )
   {
      // Locate the RTT node from the pass index and traverse the geometry.
      ColladaRTT *pRtt = m_RTTIndexMap[ nPassIndex ] ;

      if ( pRtt )
      {
         pRtt->GetRootNode()->traverse( nv ) ;
      }
   }
}


