//=================================================================================================================================
//
//          Includes / defines / typedefs / static member variable initialization block
//
//=================================================================================================================================
#include <osg/GLExtensions>
#include <osgUtil/CullVisitor>
#include <osgUtil/UpdateVisitor>
#include <osg/StateAttribute>
#include <osg/State>
#include <osg/Uniform>

#include "ColladaRTT.h"
#include "daeRMatrixUpdUtil.h"

#include <iostream>
#include <sstream>

unsigned int ColladaRTT::m_nRTTCount = 0 ;

//=================================================================================================================================
//
//          Constructor(s) / Destructor(s) Block 
//
//=================================================================================================================================

//=================================================================================================================================
//
//          Public methods block
//
//=================================================================================================================================

void ColladaRTT::Setup()
{
   m_pTextureGeometry      = NULL ;
   m_nTexWidth             = 0 ;
   m_nTexHeight            = 0 ;
   m_bUseImage             = false ;
   m_bUseTextureRectangle  = false ;
   m_bUseHDR               = false ;
   
   m_pRoot                 = NULL ;
   m_pCamera               = NULL ;
   m_pTexture              = NULL ;
   m_pGeode                = NULL ;
   m_pTestGeom             = NULL ;
   m_pTextureGeometry      = NULL ;

   m_fClearColor.set ( 0.0, 0.0, 0.0, 0.0 ) ;
   m_RTI =  osg::Camera::FRAME_BUFFER_OBJECT ;

   // Setup the root node with a Geode.
   m_pRoot = new osg::Group ;
   m_pGeode = new osg::Geode() ;
   m_pRoot->addChild( m_pGeode ) ;

   m_bCameraSet      = false ;
   m_bTextureSet     = false ;
   m_bClearColorSet  = false ;
   m_bClearDepthSet  = false ;

   m_nMyRTTCount = m_nRTTCount;
   
   m_bMainViewLock = false;

   m_vEyePos      = osg::Vec3f ( 0.0, 180.0, 0.0 );
   m_vLookAtPos   = osg::Vec3f ( 0.0, 0.0, 0.0 );
   m_vEyeUp       = osg::Vec3f ( 0.0, 0.0, 1.0 );

   m_fZnear = 1.0f ;
   m_fZfar  = 1000.0f;
   m_fXFOV  = 45.0f ;
   m_fAspectRatio;

   m_nRTTCount++;
}

bool ColladaRTT::IsComplete()
{
   if (  m_bCameraSet &&
         m_bTextureSet )
   {
      return true; 
   }

   return false;
}

void ColladaRTT::CompleteNode()
{
   if ( IsComplete () )
   {
      // Construct the node.
      m_pCamera = new osg::Camera;
      // set view
      m_pCamera->setReferenceFrame( osg::Transform::ABSOLUTE_RF );

      if ( m_bClearColorSet && !m_bClearDepthSet )
      {
         // set up the background color and clear mask.
         m_pCamera->setClearColor( m_fClearColor );
         m_pCamera->setClearMask( GL_COLOR_BUFFER_BIT  );
      }
      
      if ( m_bClearDepthSet && !m_bClearColorSet  )
      {
         // set up the background color and clear mask.
         m_pCamera->setClearMask( GL_DEPTH_BUFFER_BIT  );
      }

      if ( m_bClearDepthSet && m_bClearColorSet  )
      {
         // set up the background color and clear mask.
         m_pCamera->setClearColor( m_fClearColor );
         m_pCamera->setClearMask( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT   );
      }

      m_pCamera->setProjectionMatrixAsPerspective( m_fXFOV, 
                                                   m_fAspectRatio, // width to height
                                                   m_fZnear, 
                                                   m_fZfar );

/*
      m_pCamera->setViewMatrixAsLookAt(   m_vEyePos,
                                          m_vLookAtPos,
                                          m_vEyeUp );
*/

      m_pCamera->setComputeNearFarMode( osgUtil::CullVisitor::DO_NOT_COMPUTE_NEAR_FAR ); 
      // set viewport
      m_pCamera->setViewport( 0, 0, m_nTexWidth, m_nTexHeight );
      // set the camera to render before the main camera.
      m_pCamera->setRenderOrder( osg::Camera::PRE_RENDER );
      // tell the camera to use OpenGL frame buffer object where supported.
      m_pCamera->setRenderTargetImplementation( m_RTI );
      // Attach the texture. The texture is really a format and sampler description. 
      m_pCamera->attach( osg::Camera::COLOR_BUFFER, m_pTexture );

      /////////////////////////////////////////////////////////////////////////////////
      // Test image output    
      osg::Image* image = new osg::Image;
      image->allocateImage( m_nTexWidth, m_nTexHeight, 1, GL_RGBA, GL_UNSIGNED_BYTE);
      // attach the image so its copied on each frame.
      m_pCamera->attach( osg::Camera::COLOR_BUFFER, image );
      // Add in the RTT count (to id the images from each pass).
		std::ostringstream sstr;
		sstr << "_" << m_nMyRTTCount ;
      std::string strIndex ( sstr.str() );

      // DEBUG: Uncomment this if you want to save the RTT pass image to disk. 
      //m_pCamera->setPostDrawCallback( new MyCameraPostDrawCallback( image, m_strTargetName + strIndex  ) );
      //m_pTexture->setImage( 0, image );

      // Rather than attach the texture directly to illustrate the texture's ability to
      // detect an image update and to subload the image onto the texture.  You needn't
      // do this when using an Image for copying to, as a seperate camera->attach(..)
      // would suffice as well, but we'll do it the long way round here just for demonstation
      // purposes (long way round meaning we'll need to copy image to main memory, then
      // copy it back to the graphics card to the texture in one frame).
      // The long way round allows us to mannually modify the copied image via the callback
      // and then let this modified image by reloaded back.

      m_pRoot->addChild( m_pCamera );
   }  
}

void ColladaRTT::SetClearColor ( float fR, float fG, float fB, float fA )
{
   // Cache values
   m_fClearColor[0] = fR;
   m_fClearColor[1] = fG;
   m_fClearColor[2] = fB;
   m_fClearColor[3] = fA;
   
   m_bClearColorSet = true ;

   if ( m_pCamera )
   {
      m_pCamera->setClearColor( m_fClearColor );

      if ( m_bClearColorSet && !m_bClearDepthSet )
      {
         // set up the background color and clear mask.
         m_pCamera->setClearMask( GL_COLOR_BUFFER_BIT  );
      }
      
      if ( !m_bClearColorSet && m_bClearDepthSet )
      {
         // set up the background color and clear mask.
         m_pCamera->setClearMask( GL_DEPTH_BUFFER_BIT  );
      }

      if ( m_bClearDepthSet && m_bClearColorSet  )
      {
         // set up the background color and clear mask.
         m_pCamera->setClearMask( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT   );
      }
   }
}

void ColladaRTT::SetClearDepth ( float fDepth )
{
   m_fClearDepth = fDepth;
   m_bClearDepthSet = true ;
   
   // FIXME: re-factor
   if ( m_pCamera )
   {
      if ( m_bClearColorSet && !m_bClearDepthSet )
      {
         // set up the background color and clear mask.
         m_pCamera->setClearMask( GL_COLOR_BUFFER_BIT  );
      }
      
      if ( !m_bClearColorSet && m_bClearDepthSet )
      {
         // set up the background color and clear mask.
         m_pCamera->setClearMask( GL_DEPTH_BUFFER_BIT  );
      }

      if ( m_bClearDepthSet && m_bClearColorSet  )
      {
         // set up the background color and clear mask.
         m_pCamera->setClearMask( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT   );
      }
   }   
}

osg::Texture * ColladaRTT::CreateTexture( unsigned nTexWidth, 
                                          unsigned nTexHeight,
                                          osg::Texture * pTexture )
{
   if ( IsComplete () )
   {
      return m_pTexture ;
   } 

   // Cache values for later use.
   m_nTexWidth = nTexWidth;
   m_nTexHeight = nTexHeight;

   if ( !pTexture )
   {
      if ( m_bUseTextureRectangle )
      {
         osg::TextureRectangle* textureRect = new osg::TextureRectangle;
         textureRect->setTextureSize( m_nTexWidth, m_nTexHeight );
         textureRect->setInternalFormat( GL_RGBA );
         textureRect->setFilter(osg::Texture2D::MIN_FILTER,osg::Texture2D::LINEAR);
         textureRect->setFilter(osg::Texture2D::MAG_FILTER,osg::Texture2D::LINEAR);
         pTexture = textureRect;
      }
      else
      {
         osg::Texture2D* texture2D = new osg::Texture2D;
         texture2D->setTextureSize( m_nTexWidth, m_nTexHeight );
         texture2D->setInternalFormat( GL_RGBA );
         texture2D->setFilter(osg::Texture2D::MIN_FILTER,osg::Texture2D::LINEAR);
         texture2D->setFilter(osg::Texture2D::MAG_FILTER,osg::Texture2D::LINEAR);
         pTexture = texture2D;
      } 

      if ( m_bUseHDR )
      {
         pTexture->setInternalFormat(GL_RGBA16F_ARB);
         pTexture->setSourceFormat(GL_RGBA);
         pTexture->setSourceType(GL_FLOAT);
      }
   }

   m_pTexture = pTexture ;
   m_bTextureSet = true ;

   CompleteNode();

   return m_pTexture ;  
}    

void ColladaRTT::AddTextureGeometry ( osg::Node * pTextureGeometry )
{
   if ( m_pTextureGeometry ) 
   {
      if ( m_pCamera ) 
      {
         m_pCamera->removeChild ( m_pTextureGeometry ) ;
      }
   }

   m_pTextureGeometry = pTextureGeometry;

   if ( m_pCamera )
   {
      m_pCamera->addChild( m_pTextureGeometry );
   }
}

bool ColladaRTT::CreateManualCamera(   osg::Vec3 & vLookAtPos,
                                       osg::Vec3 & vEyePos,
                                       osg::Vec3 & vEyeUp,
                                       float       fZnear,
                                       float       fZfar,
                                       float       fXFOV,
                                       float       fAspectRatio )
{
   if ( IsComplete () )
   {
      return true;
   } 

   // Cache these values
   m_vLookAtPos   = vLookAtPos;
   m_vEyePos      = vEyePos;
   m_vEyeUp       = vEyeUp;
   m_fZnear       = fZnear;
   m_fZfar        = fZfar;
   m_fXFOV        = fXFOV;
   m_fAspectRatio = fAspectRatio;

   m_bCameraSet = true ;

   CompleteNode();

   return true;
}

void ColladaRTT::SetMainViewLock( bool bVal )
{
   m_bMainViewLock = bVal;
}

void ColladaRTT::HandleUpdate ( osg::NodeVisitor& nv, osg::Matrixf  & masterCameraVPW )
{
   // Make the camera absolute and not reletive to the current view transform
   m_pCamera->setReferenceFrame( osg::Transform::ABSOLUTE_RF );
   m_pCamera->setViewMatrix ( masterCameraVPW );
}

//=================================================================================================================================
//
//          Protected methods block
//
//=================================================================================================================================

//=================================================================================================================================
//
//          Private methods block
//
//=================================================================================================================================



