//osgFX - Copyright (C) 2003 Marco Jez

// AMD /////////////////////////////////////////////////////////////////
// DESCRIPTION: This code is looslely based on the osgFX Technique class. 
// Extends the ColladaEffect class providing RTT and geometry
// support. 
// Copyright (c) 2008 Advanced Micro Devices, Inc. 
////////////////////////////////////////////////////////////////////////

#ifndef __COLLADA_TECHNIQUE__
#define __COLLADA_TECHNIQUE__

#include <osgFX/Export>
#include <osg/Referenced>
#include <osg/State>
#include <osg/Group>
#include <osg/NodeVisitor>

#include <vector>
#include <string>

/**
 An helper macro that defines the methods techniqueName() and 
 techniqueDescription() making them return the strings passed as parameters.
 */
#define META_Technique(name, description) \
    inline virtual const char *techniqueName() { return name; } \
    inline virtual const char *techniqueDescription() { return description; }

    class ColladaEffectNode;

    /**
     This is the base class for effect techniques. A technique represents one
     of the possible ways to implement a special effect. This base class is
     abstract, you will have to subclass your own techniques for your custom
     effects.
     Derived classes will have to implement the define_passes() method to
     configure the rendering pass(es) that make up the technique. Usually
     you will create one StateSet object for each rendering pass and then
     you'll call addPass(stateset).
     The validate() method should return true if the technique is valid within
     the current rendering context, false otherwise. The default implementation
     of validate() calls getRequiredExtensions() and tests whether all required
     extensions are supported or not, returning false if at least one extension
     is not supported.
     */
    class  ColladaTechnique: public osg::Referenced 
    {

    public:

        ColladaTechnique();

        /** get the name of this Technique */
        virtual const char *techniqueName()        { return "Default"; }

        /** get a brief description of this Technique */
        virtual const char *techniqueDescription() { return "This is the default technique"; }

        /**
         collect the GL extension strings which are required for this technique
         to work properly. This method is called from the default implementation
         of validate().
         */
        virtual void getRequiredExtensions(std::vector<std::string>&  /*extensions*/) const {};

        /**
         tests whether this technique is valid for the current rendering context.
         The default behavior is to call getRequiredExtensions() and check for
         extension availability.
         */
        virtual bool validate(osg::State& ) const;

        /** get the number of rendering passes defined in this Technique */
        virtual inline int getNumPasses() const;

        /** get the StateSet object associated to the i-th pass */
        virtual inline osg::StateSet* getPassStateSet(int i);

        /** get the const StateSet object associated to the i-th pass */
        virtual inline const osg::StateSet* getPassStateSet(int i) const;

        /** 
         traverse children with multipass if necessary. By default this method
         simply calls the protected method traverse_implementation(); you can
         override it to change the default behavior.
         Don't call this method directly as it is called by osgFX::Effect
         */
        inline virtual void traverse(osg::NodeVisitor& nv, ColladaEffectNode* fx);
        
        /** create a new pass node, add it to the technique and associate a StateSet */
        virtual void addPass( osg::StateSet* ss = 0 );
        
        // AMD Added
        void cachePass ( osg::StateSet *pState )
        {
           m_PassCache.push_back( pState ) ;
        }
        
    protected:

        ColladaTechnique(const ColladaTechnique &): osg::Referenced() {}    // copying is nonsense ;)
        
        virtual ~ColladaTechnique() {}
        ColladaTechnique &operator=(const ColladaTechnique &)  { return *this; }

        /** force rebuilding of pass nodes on next traversal */
        inline void dirtyPasses();
        
        /** optional: return a node that overrides the child node on a specified pass */
        inline virtual osg::Node* getOverrideChild(int)  { return 0; }

        /**
         traverse children with multipass if necessary. Don't call this method
         directly unless you are in a customized version of traverse().
         */
        virtual void traverse_implementation( osg::NodeVisitor& nv, ColladaEffectNode* fx );

        // AMD Added
        void define_passes()
        {     
           std::vector< osg::StateSet* >::iterator iter;

           for ( iter = m_PassCache.begin(); iter != m_PassCache.end(); ++iter ) 
           {
              addPass ( *iter );
           }        
        }

    private:
    
        typedef std::vector < osg::ref_ptr <osg::StateSet> > Pass_list;
        Pass_list _passes;
       
        // AMD Added
        std::vector< osg::StateSet*> m_PassCache;
    };

    // INLINE METHODS

    inline int ColladaTechnique::getNumPasses() const
    {
        return static_cast<int>(_passes.size());
    }

    inline osg::StateSet* ColladaTechnique::getPassStateSet(int i)
    {
        return _passes[i].get();
    }

    inline const osg::StateSet* ColladaTechnique::getPassStateSet(int i) const
    {
        return _passes[i].get();      
    }
      
    inline void ColladaTechnique::dirtyPasses()
    {
        _passes.clear();
    }
    
    inline void ColladaTechnique::traverse(osg::NodeVisitor& nv, ColladaEffectNode* fx)
    {
        traverse_implementation(nv, fx);
    }

#endif
