/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the SCEA Shared Source License, Version 1.0 (the "License"); you may not use this 
 * file except in compliance with the License. You may obtain a copy of the License at:
 * http://research.scea.com/scea_shared_source_license.html
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License 
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or 
 * implied. See the License for the specific language governing permissions and limitations under the 
 * License. 
 */

#include "daeReader.h"

#include <dae.h>
#include <dae/daeSIDResolver.h>
#include <dae/domAny.h>
#include <dom/domCOLLADA.h>
#include <dom/domProfile_COMMON.h>
#include <dom/domInstance_material.h> // AMD

//HF
#include <dom/domProfile_GLSL.h>

#include <osg/BlendColor>
#include <osg/BlendFunc>
#include <osg/Texture2D>
#include <osgDB/Registry>
#include <osgDB/ReadFile>

#include <sstream>

#include "ColladaEffectNode.h" // AMD

using namespace osgdae;

// AMD
ColladaEffectNode * daeReader::processBindMaterialUsingFX( domBind_material *bm, osg::Node *pParent )
{
   if ( bm->getTechnique_common() == NULL )
   {
      osg::notify( osg::WARN ) << "No COMMON technique for bind_material" << std::endl;
      return NULL; // AMD
   }

   domInstance_material_Array &ima = bm->getTechnique_common()->getInstance_material_array();
   size_t count = ima.getCount();
   for ( size_t i = 0; i < count; i++ )
   {
      std::string symbol = ima[i]->getSymbol();
      domMaterial *mat = daeSafeCast< domMaterial >( getElementFromURI( ima[i]->getTarget() ) );
      if ( mat == NULL ) 
      {
         osg::notify( osg::WARN ) << "Failed to locate material " << ima[i]->getTarget().getURI() << std::endl;
         continue;
      }

      // AMD 
	   // Get the instance material
	   domInstance_materialRef matRef = ima[i];

		/**
		 * HF: each object HAS to have it's own stateset, else the uniform - callbacks
		 * for WORLD-space matrices won't work
		 * However, in processProfile_GLSL you'll find another caching method
		 */

	   //check material cache if this material already exists
      // AMD
      ColladaEffectNode *pColladaEffect = processMaterialUsingFX( mat, matRef, pParent );
      if ( pColladaEffect == NULL )
      {
         continue;
      }

      m_materialEffectMap.insert( std::make_pair( mat, pColladaEffect ) );
 
      return pColladaEffect; // AMD
	}

   return NULL; // AMD
}


ColladaEffectNode *daeReader::processMaterialUsingFX( domMaterial             *  pDomMaterial, 
                                                      domInstance_materialRef    rDomMaterialRef,
                                                      osg::Node               *  pParent )
{
   domEffect *pDomEffect = daeSafeCast< domEffect >( getElementFromURI( pDomMaterial->getInstance_effect()->getUrl() ) );
   if ( pDomEffect == NULL )
   {
      osg::notify( osg::WARN ) << "Failed to locate effect " << pDomMaterial->getInstance_effect()->getUrl().getURI() << std::endl;
      return NULL;
   }

	m_pCurrentEffect = pDomEffect;
	
   // Create a new COLLADA effect based on the Effect and Materials data 
	ColladaEffectNode * pColladaEffect = NULL;
   pColladaEffect = processEffectUsingFX( pDomEffect, rDomMaterialRef, pDomMaterial, pParent );
	if ( pColladaEffect == NULL ) 
   {
		osg::notify( osg::WARN ) << "Process Effect failed" << std::endl;
		return NULL;
	}

   // Override the Effect uniforms with those of the parent Material
   //unsigned int nTechCount = pColladaEffect->getNumTechniques();
   //for ( unsigned int nTechIndex = 0 ; nTechIndex < nTechCount ; nTechIndex ++ )
   //{
   //   ColladaTechnique * pTech = pColladaEffect->getTechnique( nTechIndex );
   //   assert  ( pTech );
   //   unsigned int nNumPasses = pTech->getNumPasses() ;
   //   for ( unsigned int nPassIndex = 0 ; nPassIndex < nNumPasses; nPassIndex++ )
   //   {
   //      osg::StateSet *pSet = pTech->getPassStateSet( nPassIndex );
   //      assert ( pSet ) ;
         //OverRideUniforms ( pSet, pDomMaterial, pDomEffect );
   //   }
   //}

	return pColladaEffect;
}

#if 0 
// AMD removed
bool daeReader::OverRideUniforms ( osg::StateSet* ss, domMaterial *mat, domEffect *effect ) 
{
	//TODO: check how to use the <technique_hint> element in this context...
	//Check for all <setparam> elements in <material>
	//NOTE: EffectData->second is a shortcut to quickly lookup
	//what uniforms we have created out of <newparam>/<setparam>, 
	//so we can quickly override those values by those found in 
	//the <material><setparam> elements

	ParamToUniformMap& paramToUniforms = m_loadedEffects[effect].second;

	for ( unsigned int i=0; i < mat->getInstance_effect()->getSetparam_array().getCount(); i++ ) 
   {
		////We simply check if we set an uniform out of a newparam
		////with same sid, if yes, override the uniform
		std::string source = std::string( mat->getInstance_effect()->getSetparam_array()[i]->getRef() );

		ParamToUniformMap::iterator it = paramToUniforms.find( source );

		//If we have not found it, we've never created any uniforms, so we can skip that one
		if ( it == paramToUniforms.end() )
      {
			continue;
      }

		//Get the related uniform name
		std::string uniformName = it->second;

		//newparam was used as uniform
		//remove from current stateset, then add with new values
		osg::ref_ptr<osg::Uniform > uniformToOverride = NULL;

		//Lookup for the SID in the uniform array
		//(As this would have been set as symbol name)
		for ( osg::StateSet::UniformList::iterator it = ss->getUniformList().begin(); it != ss->getUniformList().end(); it++ )
		{
			if ( it->first == uniformName )
         {
				uniformToOverride = it->second.first;
            break ;
         }
		}
		
		if ( uniformToOverride == NULL )
      {
			continue;
      }
		
		//Save important data
		std::string uniformSymbol( uniformToOverride->getName() );
		int texUnit = -1;
		if (  uniformToOverride->getType() == osg::Uniform::SAMPLER_1D ||
			   uniformToOverride->getType() == osg::Uniform::SAMPLER_2D ||
			   uniformToOverride->getType() == osg::Uniform::SAMPLER_2D_SHADOW ||
			   uniformToOverride->getType() == osg::Uniform::SAMPLER_3D ||
			   uniformToOverride->getType() == osg::Uniform::SAMPLER_CUBE ) 
		{
			uniformToOverride->get( texUnit );
			//Remove texture of sampler from stateset
		}

		//NOTE: we HAVE to skip those float4x4 values that have 
		//a viewMatrixPickup (saved in userData of osg::Unifom of type float_mat4
		if ( uniformToOverride->getType() == osg::Uniform::FLOAT_MAT4 && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		if ( uniformToOverride->getType() == osg::Uniform::FLOAT_VEC4 && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		if ( uniformToOverride->getType() == osg::Uniform::FLOAT_VEC3 && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		if ( uniformToOverride->getType() == osg::Uniform::FLOAT && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		if ( uniformToOverride->getType() == osg::Uniform::INT && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		osg::Uniform* newUniform = NULL;

      // FIXME
      osg::Uniform *pTmpToOverride = uniformToOverride.get() ;
      ss->removeUniform( pTmpToOverride );

		//NOTE: in <material>'s <setparam> there are FX Common params only...
		//So we DON'T have to check semantics (as they are not allowed in FX Common params)
		if ( !processFxCommonParamType(  mat->getInstance_effect()->getSetparam_array()[i]->getFx_basic_type_common(), 
                                       uniformSymbol.c_str(), 
                                       newUniform) ) 
      {
			osg::notify( osg::WARN ) << "Updating uniform " << uniformSymbol << " failed " << std::endl;
			return false;
		}
	
		//Update the Effect-Data
		if ( newUniform != NULL ) 
      {
			//Check if we had a sampler...
			if (  newUniform->getType() == osg::Uniform::SAMPLER_1D ||
				   newUniform->getType() == osg::Uniform::SAMPLER_2D ||
				   newUniform->getType() == osg::Uniform::SAMPLER_2D_SHADOW ||
				   newUniform->getType() == osg::Uniform::SAMPLER_3D ||
				   newUniform->getType() == osg::Uniform::SAMPLER_CUBE ) 
			{
				//Set the tex-unit (the last one as the old one)
				newUniform->set( texUnit );
				//Add new osg::Texture to stateset
				ss->setTextureAttributeAndModes( texUnit, static_cast<osg::Texture* >( newUniform->getUserData() ) );
			}

			//Add updated uniform to stateset
         // FIXME
         //ss->addUniform( pTmpToOverride );
			ss->addUniform( newUniform );
		}
	}

   return true;
}

bool daeReader::OverRideUniforms ( osg::StateSet* ss, domMaterial *mat, domProfile_GLSL::domTechnique::domPass* pDomPass )
{
	//TODO: check how to use the <technique_hint> element in this context...
	//Check for all <setparam> elements in <material>
	//NOTE: EffectData->second is a shortcut to quickly lookup
	//what uniforms we have created out of <newparam>/<setparam>, 
	//so we can quickly override those values by those found in 
	//the <material><setparam> elements

	ParamToUniformMap& paramToUniforms = m_loadedPasses[pDomPass].second;

   // Override the SetParam array data
   unsigned int nCount = mat->getInstance_effect()->getSetparam_array().getCount();
	for ( unsigned int i=0; i < nCount; i++ ) 
   {
		////We simply check if we set an uniform out of a newparam
		////with same sid, if yes, override the uniform
		std::string source = std::string( mat->getInstance_effect()->getSetparam_array()[i]->getRef() );

		ParamToUniformMap::iterator it = paramToUniforms.find( source );

		//If we have not found it, we've never created any uniforms, so we can skip that one
		if ( it == paramToUniforms.end() )
			continue;

		//Get the related uniform name
		std::string uniformName = it->second;

		//newparam was used as uniform
		//remove from current stateset, then add with new values
		osg::ref_ptr<osg::Uniform > uniformToOverride = NULL;

		//Lookup for the SID in the uniform array
		//(As this would have been set as symbol name)
		for ( osg::StateSet::UniformList::iterator it = ss->getUniformList().begin(); it != ss->getUniformList().end(); it++ )
		{
			if ( it->first == uniformName )
				uniformToOverride = it->second.first;
		}
		
		if ( uniformToOverride == NULL )
			continue;
		
		//Save important data
		std::string uniformSymbol( uniformToOverride->getName() );
		int texUnit = -1;
		if (  uniformToOverride->getType() == osg::Uniform::SAMPLER_1D ||
			   uniformToOverride->getType() == osg::Uniform::SAMPLER_2D ||
			   uniformToOverride->getType() == osg::Uniform::SAMPLER_2D_SHADOW ||
			   uniformToOverride->getType() == osg::Uniform::SAMPLER_3D ||
			   uniformToOverride->getType() == osg::Uniform::SAMPLER_CUBE ) 
		{
			uniformToOverride->get( texUnit );
			//Remove texture of sampler from stateset
		}

		//NOTE: we HAVE to skip those float4x4 values that have 
		//a viewMatrixPickup (saved in userData of osg::Unifom of type float_mat4
		if ( uniformToOverride->getType() == osg::Uniform::FLOAT_MAT4 && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		if ( uniformToOverride->getType() == osg::Uniform::FLOAT_VEC4 && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		if ( uniformToOverride->getType() == osg::Uniform::FLOAT_VEC3 && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		if ( uniformToOverride->getType() == osg::Uniform::FLOAT && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		if ( uniformToOverride->getType() == osg::Uniform::INT && 
			( uniformToOverride->getUserData() != NULL || uniformToOverride->getUpdateCallback() != NULL ) )
			continue;

		osg::Uniform* newUniform = NULL;
      osg::Uniform* pTemp = uniformToOverride.get();
      ss->removeUniform( pTemp );

		//NOTE: in <material>'s <setparam> there are FX Common params only...
		//So we DON'T have to check semantics (as they are not allowed in FX Common params)
		if ( !processFxCommonParamType(  mat->getInstance_effect()->getSetparam_array()[i]->getFx_basic_type_common(), 
                                       uniformSymbol.c_str(), 
                                       newUniform) ) 
      {
			osg::notify( osg::WARN ) << "Updating uniform " << uniformSymbol << " failed " << std::endl;
			return false;
		}
	
		//Update the Effect-Data
		if ( newUniform != NULL ) 
      {
			//Check if we had a sampler...
			if (  newUniform->getType() == osg::Uniform::SAMPLER_1D ||
				   newUniform->getType() == osg::Uniform::SAMPLER_2D ||
				   newUniform->getType() == osg::Uniform::SAMPLER_2D_SHADOW ||
				   newUniform->getType() == osg::Uniform::SAMPLER_3D ||
				   newUniform->getType() == osg::Uniform::SAMPLER_CUBE ) 
			{
				//Set the tex-unit (the last one as the old one)
				newUniform->set( texUnit );
				//Add new osg::Texture to stateset
				ss->setTextureAttributeAndModes( texUnit, static_cast<osg::Texture* >( newUniform->getUserData() ) );
			}

			//Add updated uniform to stateset
			ss->addUniform( newUniform );
		}
	}

   return true;
}
#endif

// AMD change
ColladaEffectNode *daeReader::processEffectUsingFX(   domEffect                  *pDomEffect, 
                                                      domInstance_materialRef    rDomMaterialRef,
                                                      domMaterial                *pDomMaterial,
                                                      osg::Node                  *pParent )
{
   bool hasCOMMON = false;
	//HF
	bool hasGLSL = false;
   osg::StateSet        *ss = NULL;
   ColladaEffectNode *pColladaEffect = NULL;

   for ( size_t i = 0; i < pDomEffect->getFx_profile_abstract_array().getCount(); i++ )
   {
		//HF
      /*
      // AMD removed
      domProfile_COMMON *pc = daeSafeCast< domProfile_COMMON >( effect->getFx_profile_abstract_array()[i] );
      if ( pc != NULL )
      {
	      if ( hasCOMMON )
	      {
		       osg::notify( osg::WARN ) << "Effect already has a profile_COMMON. Skipping this one" << std::endl;
		       continue;
	      }
	      ss = processProfileCOMMON( pc );
	      hasCOMMON = true;
	      continue;
      }
      */

		// AMD
		domProfile_GLSL *pGlsl = daeSafeCast< domProfile_GLSL >( pDomEffect->getFx_profile_abstract_array()[i] );
		//HF
      if ( pGlsl != NULL ) 
		{
			if ( hasGLSL ) 
                        {
				osg::notify( osg::WARN ) << "Effect already has s profile_GLSL. Skipping this one." << std::endl;
				continue;
			}
		 
         // AMD  
         pColladaEffect = processProfileGLSLUsingFX( pGlsl, rDomMaterialRef, pDomMaterial, pDomEffect, pParent );
         if ( !pColladaEffect )
         {
            osg::notify( osg::WARN ) << "Effect was not created " << pDomEffect->getFx_profile_abstract_array()[i]->getTypeName() << std::endl;
            return NULL;
         }
			continue;
		}
		
      osg::notify( osg::WARN ) << "unsupported effect profile " << pDomEffect->getFx_profile_abstract_array()[i]->getTypeName() << std::endl;
   }

   return pColladaEffect;
}

osg::StateSet *daeReader::processProfileCOMMON( domProfile_COMMON *pc )
{
	 osg::StateSet *ss = new osg::StateSet();
	 
	 domProfile_COMMON::domTechnique *teq = pc->getTechnique();

	 domProfile_COMMON::domTechnique::domConstant *c = teq->getConstant();
	 domProfile_COMMON::domTechnique::domLambert *l = teq->getLambert();
	 domProfile_COMMON::domTechnique::domPhong *p = teq->getPhong();
	 domProfile_COMMON::domTechnique::domBlinn *b = teq->getBlinn();

	 ss->setMode( GL_CULL_FACE, GL_TRUE );
	 //ss->setMode( GL_LIGHTING, GL_FALSE );

	 osg::ref_ptr< osg::Material > mat = new osg::Material();
	 bool insertMat = false;
	 if ( b != NULL )
	 {
		  bool tmp;
		  tmp = processColorOrTextureType( b->getEmission(), osg::Material::EMISSION, mat.get() );
		  insertMat = insertMat || tmp;
		  
		  tmp = processColorOrTextureType( b->getAmbient(), osg::Material::AMBIENT, mat.get() );
		  insertMat = insertMat || tmp;
		  
		  osg::StateAttribute *sa = NULL;
		  tmp = processColorOrTextureType( b->getDiffuse(), osg::Material::DIFFUSE, mat.get(), NULL, &sa );
		  insertMat = insertMat || tmp;
		  if ( sa != NULL ) 
		  {
				ss->setTextureMode( 0, GL_TEXTURE_2D, GL_TRUE );
				ss->setTextureAttribute( 0, sa );
		  }

		  tmp = processColorOrTextureType( b->getSpecular(), osg::Material::SPECULAR, mat.get(), b->getShininess() );
		  insertMat = insertMat || tmp;

		  osg::StateAttribute *sa2 = NULL;
		  sa2 = processTransparentType( b->getTransparent(), ss );
		  if ( sa2 != NULL )
		  {
				if ( sa == NULL )
				{
					 ss->setTextureMode( 0, GL_TEXTURE_2D, GL_TRUE );
					 ss->setTextureAttribute( 0, sa2 );
				}
				else
				{
					 osg::notify( osg::WARN ) << "Already have a texture in the diffuse channel" << std::endl;
				}
		  }

	 }
	 else if ( p != NULL )
	 {
		  bool tmp;
		  tmp = processColorOrTextureType( p->getEmission(), osg::Material::EMISSION, mat.get() );
		  insertMat = insertMat || tmp;
		  
		  tmp = processColorOrTextureType( p->getAmbient(), osg::Material::AMBIENT, mat.get() );
		  insertMat = insertMat || tmp;
		  
		  osg::StateAttribute *sa = NULL;
		  tmp = processColorOrTextureType( p->getDiffuse(), osg::Material::DIFFUSE, mat.get(), NULL, &sa );
		  insertMat = insertMat || tmp;
		  if ( sa != NULL ) 
		  {
				ss->setTextureMode( 0, GL_TEXTURE_2D, GL_TRUE );
				ss->setTextureAttribute( 0, sa );
		  }
		  
		  tmp = processColorOrTextureType( p->getSpecular(), osg::Material::SPECULAR, mat.get(), p->getShininess() );
		  insertMat = insertMat || tmp;

		  osg::StateAttribute *sa2 = NULL;
		  sa2 = processTransparentType( p->getTransparent(), ss );
		  if ( sa2 != NULL )
		  {
				if ( sa == NULL )
				{
					 ss->setTextureMode( 0, GL_TEXTURE_2D, GL_TRUE );
					 ss->setTextureAttribute( 0, sa2 );
				}
				else
				{
					 osg::notify( osg::WARN ) << "Already have a texture in the diffuse channel" << std::endl;
				}
		  }

	 }
	 else if ( l != NULL )
	 {
		  bool tmp;
		  tmp = processColorOrTextureType( l->getEmission(), osg::Material::EMISSION, mat.get() );
		  insertMat = insertMat || tmp;
		  
		  tmp = processColorOrTextureType( l->getAmbient(), osg::Material::AMBIENT, mat.get() );
		  insertMat = insertMat || tmp;
		  
		  osg::StateAttribute *sa = NULL;
		  tmp = processColorOrTextureType( l->getDiffuse(), osg::Material::DIFFUSE, mat.get(), NULL, &sa );
		  insertMat = insertMat || tmp;
		  if ( sa != NULL ) 
		  {
				ss->setTextureMode( 0, GL_TEXTURE_2D, GL_TRUE );
				ss->setTextureAttribute( 0, sa );
		  }

		  osg::StateAttribute *sa2 = NULL;
		  sa2 = processTransparentType( l->getTransparent(), ss );
		  if ( sa2 != NULL )
		  {
				if ( sa == NULL )
				{
					 ss->setTextureMode( 0, GL_TEXTURE_2D, GL_TRUE );
					 ss->setTextureAttribute( 0, sa2 );
				}
				else
				{
					 osg::notify( osg::WARN ) << "Already have a texture in the diffuse channel" << std::endl;
				}
		  }
		  
	 }
	 else if ( c != NULL )
	 {
		  insertMat = processColorOrTextureType( c->getEmission(), osg::Material::EMISSION, mat.get() );

		  osg::StateAttribute *sa2 = NULL;
		  sa2 = processTransparentType( c->getTransparent(), ss );
		  if ( sa2 != NULL )
		  {
				ss->setTextureMode( 0, GL_TEXTURE_2D, GL_TRUE );
				ss->setTextureAttribute( 0, sa2 );
		  }
	 }
	 if ( insertMat )
	 {
		  ss->setAttribute( mat.get() );
	 }

	 return ss;
}

bool daeReader::processColorOrTextureType( domCommon_color_or_texture_type *cot, 
osg::Material::ColorMode channel,
osg::Material *mat,
domCommon_float_or_param_type *fop,
osg::StateAttribute **sa )
{
	 if ( cot == NULL )
	 {
		  return false;
	 }
	 bool retVal = false;
	 //osg::StateAttribute *sa = NULL;
	 //TODO: Make all channels process <param ref=""> type of value
	 if ( channel == osg::Material::EMISSION )
	 {
		  if ( cot->getColor() != NULL )
		  {
				domFloat4 &f4 = cot->getColor()->getValue();
				mat->setEmission( osg::Material::FRONT_AND_BACK, osg::Vec4( f4[0], f4[1], f4[2], f4[3] ) );
				retVal = true;
		  }
		  else
		  {
				osg::notify( osg::WARN ) << "Currently no support for <param> or <texture> in Emission channel " << std::endl;
		  }       
	 }
	 else if ( channel == osg::Material::AMBIENT )
	 {
		  if ( cot->getColor() != NULL )
		  {
				domFloat4 &f4 = cot->getColor()->getValue();
				mat->setAmbient( osg::Material::FRONT_AND_BACK, osg::Vec4( f4[0], f4[1], f4[2], f4[3] ) );
				retVal = true;
		  }
		  else
		  {
				osg::notify( osg::WARN ) << "Currently no support for <param> or <texture> in Ambient channel " << std::endl;
		  }
	 }
	 else if ( channel == osg::Material::DIFFUSE )
	 {
		  if ( cot->getColor() != NULL )
		  {
				domFloat4 &f4 = cot->getColor()->getValue();
				mat->setDiffuse( osg::Material::FRONT_AND_BACK, osg::Vec4( f4[0], f4[1], f4[2], f4[3] ) );
				retVal = true;
		  }
		  else if ( cot->getTexture() != NULL && sa != NULL )
		  {
				*sa = processTexture( cot->getTexture() );
				domExtra *extra = cot->getTexture()->getExtra();
				if ( extra != NULL && extra->getType() != NULL && 
					  strcmp( extra->getType(), "color" ) == 0 )
				{
					 //the extra data for osg. Diffuse color can happen with a texture.
					 for ( unsigned int i = 0; i < extra->getTechnique_array().getCount(); i++ )
					 {
						  domTechnique *teq = extra->getTechnique_array()[i];
						  if ( strcmp( teq->getProfile(), "SCEI" ) == 0 )
						  {
								osg::Vec4 col;
								domAny *dcol = (domAny*)(daeElement*)teq->getContents()[0];
								std::istringstream diffuse_colour((const char *)dcol->getValue());
								diffuse_colour >> col.r() >> col.g() >> col.b() >> col.a();
								mat->setDiffuse( osg::Material::FRONT_AND_BACK, col );
								retVal = true;
								break;
						  }
					 }
				}
		  }
		  else
		  {
				osg::notify( osg::WARN ) << "Currently no support for <param> in Diffuse channel " << std::endl;
		  }
	 }
	 else if ( channel == osg::Material::SPECULAR )
	 {
		  if ( cot->getColor() != NULL )
		  {
				domFloat4 &f4 = cot->getColor()->getValue();
				mat->setSpecular( osg::Material::FRONT_AND_BACK, osg::Vec4( f4[0], f4[1], f4[2], f4[3] ) );
				retVal = true;
		  }
		  else
		  {
				osg::notify( osg::WARN ) << "Currently no support for <param> or <texture> in Specular channel " << std::endl;
		  }
		  if ( fop != NULL && fop->getFloat() != NULL )
		  {
				mat->setShininess( osg::Material::FRONT_AND_BACK, fop->getFloat()->getValue() );
				retVal = true;
		  }
	 }

	 return retVal;
}

osg::StateAttribute *daeReader::processTexture( domCommon_color_or_texture_type_complexType::domTexture *tex )
{
	 //find the newparam for the sampler based on the texture attribute
	 domFx_sampler2D_common *sampler = NULL;
	 domFx_surface_common *surface = NULL;
	 domImage *dImg = NULL;

	 std::string target = std::string("./") + std::string(tex->getTexture());
	 osg::notify(osg::NOTICE)<<"processTexture("<<target<<")"<<std::endl;
	 
	 daeSIDResolver res1( m_pCurrentEffect, target.c_str() );
	 daeElement *el = res1.getElement();

	 if ( el == NULL )
	 {
		  osg::notify( osg::WARN ) << "Could not locate newparam for texture sampler2D" << tex->getTexture() << std::endl;
		  osg::notify( osg::WARN ) << "Checking if data does incorrect linking straight to the image" << std::endl;
		  m_pDae->getDatabase()->getElement( (daeElement**)&dImg, 0, tex->getTexture(), "image" );
		  if ( dImg != NULL )
		  {
				osg::notify( osg::WARN ) << "Direct image link found. Data is incorrect but will continue to load texture" << std::endl;
		  }
	 }
	 else
	 {
		  domCommon_newparam_type *cnp = daeSafeCast< domCommon_newparam_type >( el ); 
		  domFx_newparam_common *npc = daeSafeCast< domFx_newparam_common >( el );

		  if ( cnp != NULL )
		  {
				sampler = cnp->getSampler2D();
		  }
		  else if ( npc != NULL )
		  {
				sampler = npc->getFx_basic_type_common()->getSampler2D();
		  }

		  if ( sampler == NULL )
		  {
				osg::notify( osg::WARN ) << "Wrong newparam type. Expected sampler2D" << std::endl;
				return NULL;
		  }

		  //find the newparam for the surface based on the sampler2D->source value
		  target = std::string("./") + std::string( sampler->getSource()->getValue() );
		  daeSIDResolver res2( m_pCurrentEffect, target.c_str() );
		  el = res2.getElement();
		  if ( el == NULL )
		  {
				osg::notify( osg::WARN ) << "Could not locate newparam for source " << sampler->getSource()->getValue() << std::endl;
				return NULL;
		  }
		  cnp = daeSafeCast< domCommon_newparam_type >( el ); 
		  npc = daeSafeCast< domFx_newparam_common >( el );

		  if ( cnp != NULL )
		  {
				surface = cnp->getSurface();
		  }
		  else if ( npc != NULL )
		  {
				surface = npc->getFx_basic_type_common()->getSurface();
		  }

		  if ( surface == NULL )
		  {
				osg::notify( osg::WARN ) << "Wrong newparam type. Expected surface" << std::endl;
				return NULL;
		  }

		  //look for the domImage based on the surface initialization stuff
		  daeIDRef &ref = surface->getFx_surface_init_common()->getInit_from_array()[0]->getValue();
		  dImg = daeSafeCast< domImage >( getElementFromIDRef( ref ) );
	 }
	 if ( dImg == NULL )
	 {
		  osg::notify( osg::WARN ) << "Could not locate image for texture" << std::endl;
		  return NULL;
	 }
	 //Got a sampler and a surface and an imaged. Time to create the texture stuff for osg
	 osg::Image *img = NULL;
	 if ( dImg->getInit_from() != NULL )
	 {
		  dImg->getInit_from()->getValue().validate();

		  if ( std::string( dImg->getInit_from()->getValue().getProtocol() ) == std::string( "file" ) )
		  {
				unsigned int bufSize = 1; //for the null char
				if ( dImg->getInit_from()->getValue().getFilepath() != NULL )
				{
					bufSize += strlen( dImg->getInit_from()->getValue().getFilepath() );
				}
				if ( dImg->getInit_from()->getValue().getFile() != NULL )
				{
					bufSize += strlen( dImg->getInit_from()->getValue().getFile() );
				}
				char *path = new char[bufSize+1];
				if ( !dImg->getInit_from()->getValue().getPath( path, bufSize ) )
				{
					osg::notify( osg::WARN ) << "Unable to get path from URI." << std::endl;
					return NULL;
				}

#ifdef WIN32
				// under windows the URI passed back from the dom includes a / i.e. /c:/path/file, so we need to jump over the leading /
				char* filename = path+1;
#else
				char* filename = path;
#endif

				img = osgDB::readImageFile( filename );

				delete [] path;

				osg::notify(osg::INFO)<<"  processTexture(..) - readImage("<<filename<<")"<<std::endl;
		  }
		  else
		  {
				osg::notify( osg::WARN ) << "Only images with a \"file\" scheme URI are supported in this version." << std::endl;
				return NULL;
		  }
	 }
	 else
	 {
		  osg::notify( osg::WARN ) << "Embedded image data is not supported in this version." << std::endl;
				return NULL;
	 }

	 osg::Texture2D *t2D = new osg::Texture2D( img );
	 //set texture parameters
	 if ( sampler != NULL )
	 {
         // WrapS
         if ( sampler->getWrap_s() != NULL )
         {
				osg::Texture::WrapMode wrap;
				switch( sampler->getWrap_s()->getValue() )
				{
				case FX_SAMPLER_WRAP_COMMON_NONE:

				case FX_SAMPLER_WRAP_COMMON_WRAP:
					 wrap = osg::Texture::REPEAT;
					 break;
				
            case FX_SAMPLER_WRAP_COMMON_MIRROR:
					 wrap = osg::Texture::MIRROR;
					 break;
				
            case FX_SAMPLER_WRAP_COMMON_CLAMP:
					 wrap = osg::Texture::CLAMP_TO_EDGE; // AMD
					 break;
				
            case FX_SAMPLER_WRAP_COMMON_BORDER:
					 wrap = osg::Texture::CLAMP_TO_BORDER;
					 break;
				
            default:
					 wrap = osg::Texture::CLAMP;
					 break;
				}
				t2D->setWrap( osg::Texture::WRAP_S, wrap );
		  }
		  else
		  {
				t2D->setWrap( osg::Texture::WRAP_S, osg::Texture::REPEAT );
		  }

        // WrapT
		  if ( sampler->getWrap_t() != NULL )
		  {
				osg::Texture::WrapMode wrap;
				switch( sampler->getWrap_t()->getValue() )
				{
				case FX_SAMPLER_WRAP_COMMON_NONE:

				case FX_SAMPLER_WRAP_COMMON_WRAP:
					 wrap = osg::Texture::REPEAT;
					 break;
				
            case FX_SAMPLER_WRAP_COMMON_MIRROR:
					 wrap = osg::Texture::MIRROR;
					 break;
				
            case FX_SAMPLER_WRAP_COMMON_CLAMP:
					 wrap = osg::Texture::CLAMP_TO_EDGE; // AMD
					 break;
				
            case FX_SAMPLER_WRAP_COMMON_BORDER:
					 wrap = osg::Texture::CLAMP_TO_BORDER;
					 break;
					 break;
								default:
					 wrap = osg::Texture::CLAMP;
					 break;
				}
				t2D->setWrap( osg::Texture::WRAP_T, wrap );
		  }
		  else
		  {
				t2D->setWrap( osg::Texture::WRAP_T, osg::Texture::REPEAT );
		  }

        // MinFilter
		  if ( sampler->getMinfilter() != NULL )
		  {
				osg::Texture::FilterMode mode;
				switch( sampler->getMinfilter()->getValue() )
				{
				
            case FX_SAMPLER_FILTER_COMMON_NEAREST:
					 mode = osg::Texture::NEAREST;
					 break;
				
            case FX_SAMPLER_FILTER_COMMON_LINEAR:
					 mode = osg::Texture::LINEAR;
					 break;
				
            case FX_SAMPLER_FILTER_COMMON_NEAREST_MIPMAP_NEAREST:
					 mode = osg::Texture::NEAREST_MIPMAP_NEAREST;
					 break;
				
            case FX_SAMPLER_FILTER_COMMON_LINEAR_MIPMAP_NEAREST:
					 mode = osg::Texture::LINEAR_MIPMAP_NEAREST;
					 break;
				case FX_SAMPLER_FILTER_COMMON_NONE:
				
            case FX_SAMPLER_FILTER_COMMON_NEAREST_MIPMAP_LINEAR:
					 mode = osg::Texture::NEAREST_MIPMAP_LINEAR;
					 break;

				case FX_SAMPLER_FILTER_COMMON_LINEAR_MIPMAP_LINEAR:
					 mode = osg::Texture::LINEAR_MIPMAP_LINEAR;
					 break;

				default:
					 mode = osg::Texture::LINEAR;
					 break;
				}
				t2D->setFilter( osg::Texture::MIN_FILTER, mode );
		  }
		  else
		  {
				t2D->setFilter( osg::Texture::MIN_FILTER, osg::Texture::NEAREST_MIPMAP_LINEAR );
		  }

        // MagFilter
		  if ( sampler->getMagfilter() != NULL )
		  {
				osg::Texture::FilterMode mode;
				switch( sampler->getMagfilter()->getValue() )
				{
				case FX_SAMPLER_FILTER_COMMON_NEAREST:
					 mode = osg::Texture::NEAREST;
					 break;
				case FX_SAMPLER_FILTER_COMMON_NONE:
				case FX_SAMPLER_FILTER_COMMON_LINEAR:
					 mode = osg::Texture::LINEAR;
					 break;
				case FX_SAMPLER_FILTER_COMMON_NEAREST_MIPMAP_NEAREST:
					 mode = osg::Texture::NEAREST_MIPMAP_NEAREST;
					 break;
				case FX_SAMPLER_FILTER_COMMON_LINEAR_MIPMAP_NEAREST:
					 mode = osg::Texture::LINEAR_MIPMAP_NEAREST;
					 break;
				case FX_SAMPLER_FILTER_COMMON_NEAREST_MIPMAP_LINEAR:
					 mode = osg::Texture::NEAREST_MIPMAP_LINEAR;
					 break;
				case FX_SAMPLER_FILTER_COMMON_LINEAR_MIPMAP_LINEAR:
					 mode = osg::Texture::LINEAR_MIPMAP_LINEAR;
					 break;
				default:
					 mode = osg::Texture::LINEAR;
					 break;
				}
				t2D->setFilter( osg::Texture::MAG_FILTER, mode );
		  }
		  else
		  {
				t2D->setFilter( osg::Texture::MAG_FILTER, osg::Texture::LINEAR );
		  }

        // BorderColor
		  if ( sampler->getBorder_color() != NULL )
		  {
				const domFloat4 &col = sampler->getBorder_color()->getValue();
				t2D->setBorderColor( osg::Vec4( col[0], col[1], col[2], col[3] ) );
		  }
	 }
	 else 
	 {
		  t2D->setWrap( osg::Texture::WRAP_S, osg::Texture::REPEAT );
		  t2D->setWrap( osg::Texture::WRAP_T, osg::Texture::REPEAT );
		  t2D->setFilter( osg::Texture::MIN_FILTER, osg::Texture::NEAREST_MIPMAP_LINEAR );
		  t2D->setFilter( osg::Texture::MAG_FILTER, osg::Texture::LINEAR );
	 }

	 return t2D;
}

osg::StateAttribute *daeReader::processTransparentType( domCommon_transparent_type *ctt, osg::StateSet *ss )
{
	 if ( ctt == NULL )
	 {
		  return false;
	 }
//  bool retVal = false;
	 osg::StateAttribute *sa = NULL;
	 if ( ctt->getColor() != NULL )
	 {
		  domFloat4 &f4 = ctt->getColor()->getValue();
		  //##compliant with OSG 1.0 API
		  osg::BlendColor *bc = new osg::BlendColor();
		  bc->setConstantColor(osg::Vec4( f4[0], f4[1], f4[2], f4[3] ));
		  ss->setAttribute( bc );
		  ss->setMode( GL_BLEND, GL_TRUE );
		  ss->setRenderingHint( osg::StateSet::TRANSPARENT_BIN );
		  ss->setRenderBinDetails( 10, "DepthSortedBin" );
	 }
	 else if ( ctt->getTexture() != NULL )
	 {
		  sa = processTexture( ctt->getTexture() );
		  osg::BlendFunc *bf = new osg::BlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		  ss->setAttribute( bf );
		  ss->setMode( GL_BLEND, GL_TRUE );
		  ss->setRenderingHint( osg::StateSet::TRANSPARENT_BIN );
		  ss->setRenderBinDetails( 10, "DepthSortedBin" );
	 }
	 return sa;
}


