#include "daeReader.h"

#include <osg/StateAttribute>
#include <osg/Uniform>

// AMD
#ifndef M_PI
# define M_PI       3.14159265358979323846  /* pi */
#endif


namespace osgdae
{
	const unsigned int NO_VALID_CONTEXT = 0xffffffff;

	// a state attribute class that grabs the initial inverse view matrix
	// and sends it to a Uniform.
	// NOTE: Only the  first context is considered
	// This Extractor is heavily based osgFX - ViewMatrixExtractor implementation

   class ViewMatrixExtractor: public osg::StateAttribute 
   {
	public:

		enum Transformation {
			NONE, 
			INVERSE, 
			TRANSPOSE, 
			INVERSETRANSPOSE
		};

		ViewMatrixExtractor()
			:    osg::StateAttribute(),
			_uniform(0),
			_first_context(NO_VALID_CONTEXT),
			_trafoType(NONE)
		{
		}

		ViewMatrixExtractor(const ViewMatrixExtractor& copy, const osg::CopyOp& copyop)
			:    osg::StateAttribute(copy, copyop),
			_uniform(static_cast<osg::Uniform* >(copyop(copy._uniform.get()))),
			_first_context(NO_VALID_CONTEXT),
			_trafoType(copy._trafoType)
		{
		}

		ViewMatrixExtractor(osg::Uniform* uniform, Transformation trafoType = NONE)
			:    osg::StateAttribute(),
			_uniform(uniform),
			_first_context(NO_VALID_CONTEXT),
			_trafoType(trafoType)
		{
		}

		//NOTE: we didn't want to use another type out of osgcore, this works, 
		//but is not correct, as this custom stateattribute should have it's own 
		//enum!!!
      META_StateAttribute( osgdae, ViewMatrixExtractor, VIEWMATRIXEXTRACTOR );

		int compare(const osg::StateAttribute& sa) const
		{
			COMPARE_StateAttribute_Types(ViewMatrixExtractor, sa);
			if (_uniform.get() != rhs._uniform.get()) return -1;
			if (_trafoType != rhs._trafoType) return -1;
			return 0;
		}

		void apply(osg::State& state) const
		{
			if (_uniform == NULL)
				return;

			if (_uniform->getType() != osg::Uniform::FLOAT_MAT4) 
         {
				osg::notify( osg::WARN ) << "MATRIX4x4 SEMANTIC UPDATE: uniform has wrong datatype." << std::endl;
				return;
			}

			if (_first_context == NO_VALID_CONTEXT) 
         {
				_first_context = state.getContextID();
			}

			if ( state.getContextID() == _first_context ) 
         {
				osg::Matrix M;
				if (_trafoType == NONE) 
            {
					M = state.getModelViewMatrix();
				} 
            else if (_trafoType == INVERSE) 
            {
					M = state.getInitialInverseViewMatrix();
				} 
            else if (_trafoType == TRANSPOSE) 
            {
					//CONTINUE to correctly access the matrices...
					M.set( 
						state.getInitialViewMatrix()(0, 0), state.getInitialViewMatrix()(1, 0), state.getInitialViewMatrix()(2, 0), state.getInitialViewMatrix()(3, 0), 
						state.getInitialViewMatrix()(0, 1), state.getInitialViewMatrix()(1, 1), state.getInitialViewMatrix()(2, 1), state.getInitialViewMatrix()(3, 1), 
						state.getInitialViewMatrix()(0, 2), state.getInitialViewMatrix()(1, 2), state.getInitialViewMatrix()(2, 2), state.getInitialViewMatrix()(3, 2), 
						state.getInitialViewMatrix()(0, 3), state.getInitialViewMatrix()(1, 3), state.getInitialViewMatrix()(2, 3), state.getInitialViewMatrix()(3, 3) 
						);
				} 
            else if (_trafoType == INVERSETRANSPOSE) 
            {
					M.set( 
						state.getInitialInverseViewMatrix()(0, 0), state.getInitialInverseViewMatrix()(1, 0), state.getInitialInverseViewMatrix()(2, 0), state.getInitialInverseViewMatrix()(3, 0), 
						state.getInitialInverseViewMatrix()(0, 1), state.getInitialInverseViewMatrix()(1, 1), state.getInitialInverseViewMatrix()(2, 1), state.getInitialInverseViewMatrix()(3, 1), 
						state.getInitialInverseViewMatrix()(0, 2), state.getInitialInverseViewMatrix()(1, 2), state.getInitialInverseViewMatrix()(2, 2), state.getInitialInverseViewMatrix()(3, 2), 
						state.getInitialInverseViewMatrix()(0, 3), state.getInitialInverseViewMatrix()(1, 3), state.getInitialInverseViewMatrix()(2, 3), state.getInitialInverseViewMatrix()(3, 3) 
						);
				} 
            else 
            {
					osg::notify( osg::WARN ) << "Unexpected Transformation Type in ViewMatrixExtractor" << std::endl;
					return;
				}

				//osg::notify( osg::WARN ) << "performing viewmatrx on" << _uniform.get() << std::endl;
				//Finally set the uniform
				_uniform->set(M);
			}
		}

	private:
		mutable osg::ref_ptr<osg::Uniform> _uniform;
		mutable unsigned int _first_context;
		Transformation _trafoType;
	};

   // AMD added
   class ViewProjectionExtractor: public osg::StateAttribute 
   {
	public:

		enum Transformation {
			NONE, 
			//INVERSE, 
			TRANSPOSE 
			//INVERSETRANSPOSE
		};

		ViewProjectionExtractor()
			:    osg::StateAttribute(),
			_uniform(0),
			_first_context(NO_VALID_CONTEXT),
			_trafoType(NONE)
		{
		}

		ViewProjectionExtractor(const ViewProjectionExtractor& copy, const osg::CopyOp& copyop)
			:    osg::StateAttribute(copy, copyop),
			_uniform(static_cast<osg::Uniform* >(copyop(copy._uniform.get()))),
			_first_context(NO_VALID_CONTEXT),
			_trafoType(copy._trafoType)
		{
		}

		ViewProjectionExtractor(osg::Uniform* uniform, Transformation trafoType = NONE)
			:    osg::StateAttribute(),
			_uniform(uniform),
			_first_context(NO_VALID_CONTEXT),
			_trafoType(trafoType)
		{
		}

		//NOTE: we didn't want to use another type out of osgcore, this works, 
		//but is not correct, as this custom stateattribute should have it's own 
		//enum!!!
		//META_StateAttribute(osgdae::daeReader, ViewMatrixExtractor, VIEWMATRIXEXTRACTOR);
		META_StateAttribute( osgFX, ViewProjectionExtractor, VIEWMATRIXEXTRACTOR );

		int compare( const osg::StateAttribute& sa ) const
		{
			COMPARE_StateAttribute_Types(ViewProjectionExtractor, sa);
			if (_uniform.get() != rhs._uniform.get()) return -1;
			if (_trafoType != rhs._trafoType) return -1;
			return 0;
		}

		void apply( osg::State& state ) const
		{
			if ( _uniform == NULL )
				return;

			if ( _uniform->getType() != osg::Uniform::FLOAT_MAT4 ) 
         {
				osg::notify( osg::WARN ) << "MATRIX4x4 SEMANTIC UPDATE: uniform has wrong datatype." << std::endl;
				return;
			}

			if ( _first_context == NO_VALID_CONTEXT ) 
         {
				_first_context = state.getContextID();
			}

			if ( state.getContextID() == _first_context ) 
         {
				osg::Matrix M;
				if ( _trafoType == NONE ) 
            {
					M = state.getProjectionMatrix();
				} 
            else if ( _trafoType == TRANSPOSE ) 
            {
					M.set( 
						state.getProjectionMatrix()(0, 0), state.getProjectionMatrix()(1, 0), state.getProjectionMatrix()(2, 0), state.getProjectionMatrix()(3, 0), 
						state.getProjectionMatrix()(0, 1), state.getProjectionMatrix()(1, 1), state.getProjectionMatrix()(2, 1), state.getProjectionMatrix()(3, 1), 
						state.getProjectionMatrix()(0, 2), state.getProjectionMatrix()(1, 2), state.getProjectionMatrix()(2, 2), state.getProjectionMatrix()(3, 2), 
						state.getProjectionMatrix()(0, 3), state.getProjectionMatrix()(1, 3), state.getProjectionMatrix()(2, 3), state.getProjectionMatrix()(3, 3) 
						);
				} 
            else 
            {
					osg::notify( osg::WARN ) << "Unexpected Transformation Type in ViewMatrixExtractor" << std::endl;
					return;
				}

				_uniform->set(M);
			}
		}

	private:
		mutable osg::ref_ptr<osg::Uniform> _uniform;
		mutable unsigned int _first_context;
		Transformation _trafoType;
	};

#if 0
   // AMD added
   class ViewPositionExtractor: public osg::StateAttribute 
   {
	public:

		ViewPositionExtractor()
			:    osg::StateAttribute(),
			_uniform(0),
			_first_context(NO_VALID_CONTEXT)
		{
		}

		ViewPositionExtractor(const ViewPositionExtractor& copy, const osg::CopyOp& copyop)
			:    osg::StateAttribute(copy, copyop),
			_uniform(static_cast<osg::Uniform* >(copyop(copy._uniform.get()))),
			_first_context(NO_VALID_CONTEXT)
		{
		}

		ViewPositionExtractor( osg::Uniform* uniform )
			:    osg::StateAttribute(),
			_uniform(uniform),
			_first_context(NO_VALID_CONTEXT)
		{
		}

		//NOTE: we didn't want to use another type out of osgcore, this works, 
		//but is not correct, as this custom stateattribute should have it's own 
		//enum!!!
		META_StateAttribute( osgdae::daeReader, ViewPositionExtractor, VIEWPOSITIONEXTRACTOR );

		int compare( const osg::StateAttribute& sa ) const
		{
			COMPARE_StateAttribute_Types( ViewPositionExtractor, sa );
			
         if (_uniform.get() != rhs._uniform.get()) 
            return -1;
			
         return 0;
		}

		void apply(osg::State& state) const
		{
			if (_uniform == NULL)
				return;

			if (_uniform->getType() != osg::Uniform::FLOAT_VEC4) 
         {
				osg::notify( osg::WARN ) << "FLOAT_VEC4 SEMANTIC UPDATE: uniform has wrong datatype." << std::endl;
				return;
			}

			if (_first_context == NO_VALID_CONTEXT) 
         {
				_first_context = state.getContextID();
			}

			if ( state.getContextID() == _first_context ) 
         {
			   osg::Matrix M;
		      M = state.getModelViewMatrix();

            osg::Vec3 eye ;
            osg::Vec3 center ;
            osg::Vec3 up ;
            float lookDistance = 1.0f ;
            
            M.getLookAt( eye, center, up, lookDistance ); 
            //printf ( "Eye at: %2f %2f %2f\n", eye.x(), eye.y(), eye.z() );

			   //Finally set the uniform
            osg::Vec4 vec( eye.x(), eye.y(), eye.z(), 1.0f );
			   _uniform->set( vec );
			}
		}

	private:
		mutable osg::ref_ptr<osg::Uniform> _uniform;
		mutable unsigned int _first_context;

	};

   // AMD added
   class ViewDirectionExtractor: public osg::StateAttribute 
   {
	public:

		ViewDirectionExtractor()
			:    osg::StateAttribute(),
			_uniform(0),
			_first_context(NO_VALID_CONTEXT)
		{
		}

		ViewDirectionExtractor(const ViewDirectionExtractor& copy, const osg::CopyOp& copyop)
			:    osg::StateAttribute(copy, copyop),
			_uniform(static_cast<osg::Uniform* >(copyop(copy._uniform.get()))),
			_first_context(NO_VALID_CONTEXT)
		{
		}

		ViewDirectionExtractor( osg::Uniform* uniform )
			:    osg::StateAttribute(),
			_uniform(uniform),
			_first_context(NO_VALID_CONTEXT)
		{
		}

		//NOTE: we didn't want to use another type out of osgcore, this works, 
		//but is not correct, as this custom stateattribute should have it's own 
		//enum!!!
		//META_StateAttribute( osgdae::daeReader, ViewDirectionExtractor, VIEWDIRECTIONEXTRACTOR );
		META_StateAttribute( osgFX, ViewDirectionExtractor, VIEWDIRECTIONEXTRACTOR );

		int compare( const osg::StateAttribute& sa ) const
		{
			COMPARE_StateAttribute_Types( ViewDirectionExtractor, sa );
			
         if (_uniform.get() != rhs._uniform.get()) 
            return -1;
			
         return 0;
		}

		void apply(osg::State& state) const
		{
			if (_uniform == NULL)
				return;

			if (_uniform->getType() != osg::Uniform::FLOAT_VEC4) 
         {
				osg::notify( osg::WARN ) << "FLOAT_VEC4 SEMANTIC UPDATE: uniform has wrong datatype." << std::endl;
				return;
			}

			if (_first_context == NO_VALID_CONTEXT) 
         {
				_first_context = state.getContextID();
			}

			if ( state.getContextID() == _first_context ) 
         {
			   osg::Matrix M;
		      M = state.getModelViewMatrix();

            osg::Vec3 eye ;
            osg::Vec3 center ;
            osg::Vec3 up ;
            float lookDistance = 1.0f ;
            
            M.getLookAt( eye, center, up, lookDistance ); 
            //printf ( "Eye at: %2f %2f %2f\n", eye.x(), eye.y(), eye.z() );

			   //Finally set the uniform
            osg::Vec4 vec( -eye.x(), -eye.y(), -eye.z(), 1.0f );
            vec.normalize();
			   _uniform->set( vec );
			}
		}

	private:
		mutable osg::ref_ptr<osg::Uniform> _uniform;
		mutable unsigned int _first_context;

	};

   // AMD added
   class PassIndexExtractor: public osg::StateAttribute 
   {
	public:

		PassIndexExtractor()
			:    osg::StateAttribute(),
			_uniform(0),
			_first_context(NO_VALID_CONTEXT)
		{
		}

		PassIndexExtractor(const PassIndexExtractor& copy, const osg::CopyOp& copyop)
			:    osg::StateAttribute(copy, copyop),
			_uniform(static_cast<osg::Uniform* >(copyop(copy._uniform.get()))),
			_first_context(NO_VALID_CONTEXT)
		{
		}

		PassIndexExtractor( osg::Uniform* uniform )
			:    osg::StateAttribute(),
			_uniform(uniform),
			_first_context(NO_VALID_CONTEXT)
		{
		}

		//NOTE: we didn't want to use another type out of osgcore, this works, 
		//but is not correct, as this custom stateattribute should have it's own 
		//enum!!!
		META_StateAttribute( osgFX, PassIndexExtractor, PASSINDEXEXTRACTOR );

		int compare( const osg::StateAttribute& sa ) const
		{
			COMPARE_StateAttribute_Types( PassIndexExtractor, sa );
			
         if (_uniform.get() != rhs._uniform.get()) 
            return -1;
			
         return 0;
		}

		void apply(osg::State& state) const
		{
			if (_uniform == NULL)
				return;

			if (_uniform->getType() != osg::Uniform::FLOAT) 
         {
				osg::notify( osg::WARN ) << "FLOAT SEMANTIC UPDATE: uniform has wrong datatype." << std::endl;
				return;
			}

			if (_first_context == NO_VALID_CONTEXT) 
         {
				_first_context = state.getContextID();
			}

			if ( state.getContextID() == _first_context ) 
         {
            /* FIXME test code commented out.
            float fCurrentPassIndex = 0.0f;
            _uniform->get( fCurrentPassIndex );
            fCurrentPassIndex+=1.0f;

            if ( fCurrentPassIndex >15.0f )
            {
               fCurrentPassIndex = 0.0f;
            }
            _uniform->set( fCurrentPassIndex );
            */
			}
		}

	private:
		mutable osg::ref_ptr<osg::Uniform> _uniform;
		mutable unsigned int _first_context;
	};
#endif

}


//Update Callback for WORLD, WORLDINVERSE, WORLDINVERSETRANSPOSE float4x4
//This is very useful for Lighting in World-space
//WATCH OUT: if using those callbacks, the uniform mustn't be attached to 
//another geometry, via same statesets
namespace osgdae 
{
	class worldMatrixCallback : public osg::Uniform::Callback
	{
	public: 

		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if ( uniform == NULL ) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			if ( uniform->getType() != osg::Uniform::FLOAT_MAT4 ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix update" << std::endl;
				return;
			}
			uniform->set( nv->getNodePath().back()->getWorldMatrices()[0] );
		}
	};

	class worldMatrixInverseCallback : public osg::Uniform::Callback
	{
	public: 
		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			if (uniform->getType() != osg::Uniform::FLOAT_MAT4) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix-inverse update" << std::endl;
				return;
			}
			//Setting Inverse World Matrix (all Transformation in path, except viewing matrix)
			uniform->set( osg::Matrixf::inverse( nv->getNodePath().back()->getWorldMatrices()[0] ));
		}
	};

   // AMD added
   class worldMatrixTransposeCallback : public osg::Uniform::Callback
	{
	public: 
		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			if ( uniform->getType() != osg::Uniform::FLOAT_MAT4 ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix-inverse-transpose update" << std::endl;
				return;
			}
			//Setting World Matrix (all Transformation in path, except viewing matrix)
			osg::Matrix worldM = nv->getNodePath().back()->getWorldMatrices()[0] ;
			osg::Matrix M;
			M.set( 
				worldM(0,0), worldM(1, 0), worldM(2, 0), worldM(3, 0), 
				worldM(0,1), worldM(1, 1), worldM(2, 1), worldM(3, 1), 
				worldM(0,2), worldM(1, 2), worldM(2, 2), worldM(3, 2), 
				worldM(0,3), worldM(1, 3), worldM(2, 3), worldM(3, 3) 
				);
			uniform->set(M);
		}
	};

	class worldMatrixInverseTransposeCallback : public osg::Uniform::Callback
	{
	public: 
		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			if (uniform->getType() != osg::Uniform::FLOAT_MAT4) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix-inverse-transpose update" << std::endl;
				return;
			}
			//Setting World Matrix (all Transformation in path, except viewing matrix)
			osg::Matrix worldM = osg::Matrixf::inverse( nv->getNodePath().back()->getWorldMatrices()[0] );
			osg::Matrix M;
			M.set( 
				worldM(0,0), worldM(1, 0), worldM(2, 0), worldM(3, 0), 
				worldM(0,1), worldM(1, 1), worldM(2, 1), worldM(3, 1), 
				worldM(0,2), worldM(1, 2), worldM(2, 2), worldM(3, 2), 
				worldM(0,3), worldM(1, 3), worldM(2, 3), worldM(3, 3) 
				);
			uniform->set(M);
		}
	};

   // AMD added
   class viewPositionCallback : public osg::Uniform::Callback
	{
	public: 

		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			
         if (uniform->getType() != osg::Uniform::FLOAT_VEC4) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for vec4 update" << std::endl;
				return;
			}	

	      osg::Matrix M;
		   M = nv->getNodePath().back()->getWorldMatrices()[0];
         osg::Vec3 eye ;
         osg::Vec3 center ;
         osg::Vec3 up ;
         float lookDistance = 1.0f ;
         M.getLookAt(eye,center,up,lookDistance); 
         osg::Vec4 vec( eye, 1.0f );
			uniform->set( vec );         
		}
	};

   // AMD added
	class Time_ZERO_X_Callback : public osg::Uniform::Callback
	{
	public: 

		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			
         if ( uniform->getType() != osg::Uniform::FLOAT ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix update" << std::endl;
				return;
			}
			
         float fTime = nv->getFrameStamp()->getReferenceTime();
         uniform->set( fTime );
		}
	};

   // AMD added
	class Time_ZERO_2PI_Callback : public osg::Uniform::Callback
	{
	public: 

      Time_ZERO_2PI_Callback::Time_ZERO_2PI_Callback( float fTimeCycle )
         : m_fTimeCyclePeriod( fTimeCycle )
      {
      }

		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			
         if ( uniform->getType() != osg::Uniform::FLOAT ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix update" << std::endl;
				return;
			}
			
         float fTime = nv->getFrameStamp()->getSimulationTime();
         float fAngle = ( fmod( fTime, m_fTimeCyclePeriod ) / m_fTimeCyclePeriod ) * 2.0f * M_PI;
         uniform->set( fAngle );
		}

   private:

      float m_fTimeCyclePeriod;
	};

   // AMD added
	class Time_SIN_ZERO_2PI_Callback : public osg::Uniform::Callback
	{

	public: 

      Time_SIN_ZERO_2PI_Callback::Time_SIN_ZERO_2PI_Callback( float fTimeCycle )
         : m_fTimeCyclePeriod( fTimeCycle )
      {
      }

		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			
         if ( uniform->getType() != osg::Uniform::FLOAT ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix update" << std::endl;
				return;
			}
			
         float fTime = nv->getFrameStamp()->getSimulationTime();
         float fAngle = ( fmod( fTime, m_fTimeCyclePeriod ) / m_fTimeCyclePeriod ) * 2.0f * M_PI;
         uniform->set( sin(-fAngle) );
		}

   private:

      float m_fTimeCyclePeriod;
	};

   // AMD added
	class Time_COS_ZERO_2PI_Callback : public osg::Uniform::Callback
	{
	public: 

      Time_COS_ZERO_2PI_Callback::Time_COS_ZERO_2PI_Callback( float fTimeCycle )
         : m_fTimeCyclePeriod( fTimeCycle )
      {
      }

		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			
         if ( uniform->getType() != osg::Uniform::FLOAT ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix update" << std::endl;
				return;
			}
			
         float fTime = nv->getFrameStamp()->getSimulationTime();
         float fAngle = ( fmod( fTime, m_fTimeCyclePeriod ) / m_fTimeCyclePeriod ) * 2.0f * M_PI;
         uniform->set( cos(-fAngle) );
		}

   private:

      float m_fTimeCyclePeriod;
	};


   // AMD added
	class Time_SIN_ZERO_X_Callback : public osg::Uniform::Callback
	{
	public: 

      Time_SIN_ZERO_X_Callback::Time_SIN_ZERO_X_Callback()
      {
      }

		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			
         if ( uniform->getType() != osg::Uniform::FLOAT ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix update" << std::endl;
				return;
			}
		
         float fTime = nv->getFrameStamp()->getSimulationTime();
         uniform->set( sin( fTime ) );
		}
	};


   // AMD added
	class Time_COS_ZERO_X_Callback : public osg::Uniform::Callback
	{
	public: 

      Time_COS_ZERO_X_Callback::Time_COS_ZERO_X_Callback()
      {
      }

		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			
         if ( uniform->getType() != osg::Uniform::FLOAT ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix update" << std::endl;
				return;
			}
		
         float fTime = nv->getFrameStamp()->getSimulationTime();
         uniform->set( cos(fTime) );
		}
	};

   // AMD added
   class Time_ZERO_ONE_Callback : public osg::Uniform::Callback
	{
	public: 

      Time_ZERO_ONE_Callback::Time_ZERO_ONE_Callback( float fTimeCycle )
         : m_fTimeCyclePeriod( fTimeCycle )
      {
      }
      
		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			
         if ( uniform->getType() != osg::Uniform::FLOAT ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix update" << std::endl;
				return;
			}
			
         float fTime = nv->getFrameStamp()->getSimulationTime();
         float fTime0to1 = fTime / __max( 1.0f, m_fTimeCyclePeriod );
         uniform->set( fTime0to1 );
		}

   private:

      float m_fTimeCyclePeriod;
	};

  // AMD added
  class PassFrameCallback : public osg::Uniform::Callback
	{
	public: 

      PassFrameCallback::PassFrameCallback()
      {
      }

		virtual void operator() (osg::Uniform* uniform, osg::NodeVisitor* nv) 
      {
			if (uniform == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a osg::Uniform with NULL ptr" << std::endl;
				return;
			}
			
         if ( uniform->getType() != osg::Uniform::FLOAT ) 
         {
				osg::notify( osg::WARN ) << "Wrong uniform type for world-matrix update" << std::endl;
				return;
			}

         // FIXME: test code commented out.
         //float fCurrentPassIndex = 0;
         //uniform->get( fCurrentPassIndex );
         //fCurrentPassIndex += 1.0f;

         //if ( fCurrentPassIndex > 2.0f )
         //{
         //   fCurrentPassIndex = 0.0f;
         //}

         //uniform->set( fCurrentPassIndex );
         //printf ( "PassIndex: %.2lf\n", fCurrentPassIndex ) ;

		}
	};

  // AMD added
  class NodeDoesNothingCallback : public osg::NodeCallback
	{
	public: 

      NodeDoesNothingCallback::NodeDoesNothingCallback()
      {
      }

		virtual void operator() ( osg::Node *pNode, osg::NodeVisitor* nv ) 
      {
			if (pNode == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a node that is NULL" << std::endl;
				return;
			}
			// Hey gues what, yes do nothing!

         traverse( pNode, nv );
		}
	};

  // AMD added
  class StateSetDoesNothingCallback : public osg::StateSet::Callback
	{
	public: 

      StateSetDoesNothingCallback::StateSetDoesNothingCallback()
      {
      }

		virtual void operator() ( osg::Node *pNode, osg::NodeVisitor* nv ) 
      {
			if (pNode == NULL) 
         {
				osg::notify( osg::WARN ) << "Trying update a node that is NULL" << std::endl;
				return;
			}
			// Hey gues what, yes do nothing!
		}
	};

}


