/*
* Copyright 2006 Sony Computer Entertainment Inc.
*
* Licensed under the SCEA Shared Source License, Version 1.0 (the "License"); you may not use this 
* file except in compliance with the License. You may obtain a copy of the License at:
* http://research.scea.com/scea_shared_source_license.html
*
* Unless required by applicable law or agreed to in writing, software distributed under the License 
* is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or 
* implied. See the License for the specific language governing permissions and limitations under the 
* License. 
*/

#include "daeReader.h"
#include <dae.h>
#include <dom/domCOLLADA.h>
#include <dom/domInstanceWithExtra.h>
#include "daeRMatrixUpdUtil.h" //AMD

using namespace osgdae;

daeReader::daeReader(DAE *dae_) 
: m_pDae(dae_),
m_pRootNode(NULL),
m_numlights(0),
m_pCurrentEffect(NULL),
m_geometryMap(),
m_materialMap(),
m_pDoc(NULL), //AMD
m_pCurrentColladaEffect(NULL), //AMD
m_pCurrentDomMaterial(NULL) //AMD
{
   //HF
   //NONE defaults to repeat
   m_wrapModeMap[FX_SAMPLER_WRAP_COMMON_NONE] = osg::Texture::REPEAT;
   m_wrapModeMap[FX_SAMPLER_WRAP_COMMON_WRAP] = osg::Texture::REPEAT;
   m_wrapModeMap[FX_SAMPLER_WRAP_COMMON_MIRROR] = osg::Texture::MIRROR;
   m_wrapModeMap[FX_SAMPLER_WRAP_COMMON_CLAMP] = osg::Texture::CLAMP_TO_EDGE; //AMD
   m_wrapModeMap[FX_SAMPLER_WRAP_COMMON_BORDER] = osg::Texture::CLAMP_TO_BORDER;

   //NONE defaults to nearest mipmap linear
   m_filterModeMap[FX_SAMPLER_FILTER_COMMON_NEAREST] = osg::Texture::NEAREST;
   m_filterModeMap[FX_SAMPLER_FILTER_COMMON_LINEAR] = osg::Texture::LINEAR;
   m_filterModeMap[FX_SAMPLER_FILTER_COMMON_NEAREST_MIPMAP_NEAREST] = osg::Texture::NEAREST_MIPMAP_NEAREST;
   m_filterModeMap[FX_SAMPLER_FILTER_COMMON_LINEAR_MIPMAP_NEAREST] = osg::Texture::LINEAR_MIPMAP_NEAREST;
   m_filterModeMap[FX_SAMPLER_FILTER_COMMON_NEAREST_MIPMAP_LINEAR] = osg::Texture::NEAREST_MIPMAP_LINEAR;
   m_filterModeMap[FX_SAMPLER_FILTER_COMMON_NONE] = osg::Texture::NEAREST_MIPMAP_LINEAR;
   m_filterModeMap[FX_SAMPLER_FILTER_COMMON_LINEAR_MIPMAP_LINEAR] = osg::Texture::LINEAR_MIPMAP_LINEAR;

   m_blendFuncModeMap[GL_BLEND_TYPE_CONSTANT_ALPHA] = osg::BlendFunc::CONSTANT_ALPHA;
   m_blendFuncModeMap[GL_BLEND_TYPE_CONSTANT_COLOR] = osg::BlendFunc::CONSTANT_COLOR;
   m_blendFuncModeMap[GL_BLEND_TYPE_DEST_COLOR] = osg::BlendFunc::DST_COLOR;
   m_blendFuncModeMap[GL_BLEND_TYPE_DST_ALPHA] = osg::BlendFunc::DST_ALPHA;
   m_blendFuncModeMap[GL_BLEND_TYPE_ONE] = osg::BlendFunc::ONE;
   m_blendFuncModeMap[GL_BLEND_TYPE_ONE_MINUS_CONSTANT_ALPHA] = osg::BlendFunc::ONE_MINUS_CONSTANT_ALPHA;
   m_blendFuncModeMap[GL_BLEND_TYPE_ONE_MINUS_CONSTANT_COLOR] = osg::BlendFunc::ONE_MINUS_CONSTANT_COLOR;
   m_blendFuncModeMap[GL_BLEND_TYPE_ONE_MINUS_DEST_COLOR] = osg::BlendFunc::ONE_MINUS_DST_COLOR;
   m_blendFuncModeMap[GL_BLEND_TYPE_ONE_MINUS_DST_ALPHA] = osg::BlendFunc::ONE_MINUS_DST_ALPHA;
   m_blendFuncModeMap[GL_BLEND_TYPE_ONE_MINUS_SRC_ALPHA] = osg::BlendFunc::ONE_MINUS_SRC_ALPHA;
   m_blendFuncModeMap[GL_BLEND_TYPE_ONE_MINUS_SRC_COLOR] = osg::BlendFunc::ONE_MINUS_SRC_COLOR;
   m_blendFuncModeMap[GL_BLEND_TYPE_SRC_ALPHA] = osg::BlendFunc::SRC_ALPHA;
   m_blendFuncModeMap[GL_BLEND_TYPE_SRC_ALPHA_SATURATE] = osg::BlendFunc::SRC_ALPHA_SATURATE;
   m_blendFuncModeMap[GL_BLEND_TYPE_SRC_COLOR] = osg::BlendFunc::SRC_COLOR;
   m_blendFuncModeMap[GL_BLEND_TYPE_ZERO] = osg::BlendFunc::ZERO;
}

daeReader::~daeReader()
{
}

bool daeReader::convert( const std::string &fileURI ) 
{

   std::string fURI;
   if ( fileURI[1] == ':' )
   {
      fURI = "/" + fileURI;
   }
   else
   {
      fURI = fileURI;
   }

   daeInt res = m_pDae->load( fURI.c_str() );

   if( res != DAE_OK && res != DAE_ERR_COLLECTION_ALREADY_EXISTS) 
   {
      osg::notify( osg::WARN ) << "Load failed in COLLADA DOM" << std::endl;
      return false;
   }
   osg::notify( osg::INFO ) << "URI loaded: " << fURI << std::endl;

   m_pDoc = m_pDae->getDom( fURI.c_str() );

   if ( !m_pDoc->getScene() || !m_pDoc->getScene()->getInstance_visual_scene() ) 
   {
      osg::notify( osg::WARN ) << "No scene found!" << std::endl;
      return false;
   }

   domInstanceWithExtra *ivs = m_pDoc->getScene()->getInstance_visual_scene();
   domVisual_scene *vs = daeSafeCast< domVisual_scene >( getElementFromURI( ivs->getUrl() ) );
   if ( vs == NULL ) 
   {
      osg::notify( osg::WARN ) << "Unable to locate visual scene!" << std::endl;
      return false;
   }

   m_refRootNode = processVisualScene( vs );

   return true;
}

osg::Node* daeReader::processVisualScene( domVisual_scene *scene )
{
   //osg::Node *retVal; 
   osg::Group *pRootGroup = new osg::Group(); //AMD
   pRootGroup->setUpdateCallback( new NodeDoesNothingCallback() ); //AMD

   //### do not add an empty group if there is only one
   unsigned int nbVisualSceneGroup = scene->getNode_array().getCount();

   if ( nbVisualSceneGroup == 0 )
   {
      osg::notify( osg::WARN ) << "No visual scene group found !" << std::endl;
      pRootGroup->setName("Empty Collada scene");
   }
   else if ( nbVisualSceneGroup == 1 )
   {
      processNode( scene->getNode_array()[0], pRootGroup );
      // AMD commented the following out
      //if ( node != NULL )
      //{
      //    retVal = node;
      //}
      //else
      //{
      //   retVal = new osg::Group();
      //    retVal->setName("Empty Collada scene (import failure)");
      //}
   }
   else
   { 
      pRootGroup->setName("Collada visual scene group");
      for ( size_t i = 0; i < scene->getNode_array().getCount(); i++ )
      {
         processNode( scene->getNode_array()[i], pRootGroup ); // AMD
         // AMD Commenetd out
         //if ( node != NULL )
         //{
         //  retVal->asGroup()->addChild( node );
         //}
      }
   }

   return pRootGroup;
}

void daeReader::processNode( domNode *node, osg::Node *pParentNode )
{
   // AMD This two pass approach should create the typical hierarchical stack matrix 
   // intended by the use of nested nodes.

   // The first pass of the source node's children looks for transformations. 
   // For each transformation we create a group node and add the transformation to it.
   // We attach each subsequent transformation as a child to the previous node.
   // So, 4 transformations would result in a 4-deep tree (e.g. t1->2->t3->t4).

   // The second pass of the source node's children looks for more nodes, instance_nodes, and the 
   // various instances to other data. Each of these items is processed and added as children to the 
   // end of the transform list (e.g. t1->2->t3->t4|->n1
   //                                              |->n2
   //                                              |->n3

   // AMD Make a new node that will always be the current leaf node of the tree.
   osg::Node * pNewParent = new osg::Group();
   pNewParent->setUpdateCallback( new NodeDoesNothingCallback() );

   // AMD Add to parent so that the UP chain is maintained.   
   pParentNode->asGroup()->addChild ( pNewParent );
   pNewParent->setName( node->getId() ? node->getId() : "Process Node Group" );

   // AMD This transformation pointer is passed into the re-entrant calls to processNode.
   // The children of the current node are added to the translation object, not the parent node.
   osg::Node *pGenericNode = NULL;

   // AMD Pass 1: Traverse the node children and see if there are transforms.
   size_t count = node->getContents().getCount();
   for ( size_t i = 0; i < count; i++ ) 
   {
      //I'm using daeSafeCast to check type because the pointer comparisons are a lot faster
      //than a strcmp

      domTranslate *pTranslate = daeSafeCast< domTranslate >( node->getContents()[i] );
      if ( pTranslate != NULL )
      {
         pGenericNode = processTranslate( pTranslate );
         if ( pGenericNode != NULL )
         {
            pNewParent->asGroup()->addChild( pGenericNode );
            pNewParent = pGenericNode;
         }
      }

      domRotate *pRot = daeSafeCast< domRotate >( node->getContents()[i] );
      if ( pRot != NULL ) 
      {
         pGenericNode = processRotate( pRot );
         if ( pGenericNode != NULL )
         {
            pNewParent->asGroup()->addChild( pGenericNode );
            pNewParent = pGenericNode;
         }
      }

      domScale *pScale = daeSafeCast< domScale >( node->getContents()[i] );
      if ( pScale != NULL ) 
      {
         pGenericNode = processScale( pScale );
         if ( pGenericNode != NULL )
         {
            pNewParent->asGroup()->addChild( pGenericNode );
            pNewParent = pGenericNode;
         }
      }

      domMatrix *pMatrix = daeSafeCast< domMatrix >( node->getContents()[i] );
      if ( pMatrix != NULL ) 
      {
         pGenericNode = processMatrix( pMatrix );
         if ( pGenericNode != NULL )
         {
            pNewParent->asGroup()->addChild( pGenericNode );
            pNewParent = pGenericNode;
         }
      }

      domSkew *pSkew = daeSafeCast< domSkew >( node->getContents()[i] );
      if ( pSkew != NULL ) 
      {
         pGenericNode = processSkew( pSkew );
         if ( pGenericNode != NULL )
         {
            pNewParent->asGroup()->addChild( pGenericNode );
            pNewParent = pGenericNode;
         }
      }
   }

   // AMD Pass 2: Traverse the contents again (ignoring transforms), and process each of the content nodes.
   for ( size_t i = 0; i < count; i++ ) 
   {
      domInstance_node *instn = daeSafeCast< domInstance_node >( node->getContents()[i] );
      if ( instn != NULL )
      {
         daeElement *el = getElementFromURI( instn->getUrl() );
         domNode *n = daeSafeCast< domNode >( el );
         if ( n == NULL )
         {
            osg::notify( osg::WARN ) << "Failed to instance node " << instn->getUrl().getURI() << std::endl;
         }
         processNode( n, pNewParent );
         continue;
      }

      domNode *n = daeSafeCast< domNode >( node->getContents()[i] );
      if ( n != NULL )
      {
         processNode( n, pNewParent );
         continue;
      }

      domInstance_geometry *ig = daeSafeCast< domInstance_geometry >( node->getContents()[i] );
      if ( ig != NULL )
      {
         processInstance_geometry( ig, pNewParent );
         continue;
      }

      domInstance_controller *ictrl = daeSafeCast< domInstance_controller >( node->getContents()[i] );
      if ( ictrl != NULL )
      {
         processInstance_controller( ictrl, pNewParent );
         continue;
      }

      domInstance_camera *ic = daeSafeCast< domInstance_camera >( node->getContents()[i] );
      if ( ic != NULL )
      {
         daeElement *el = getElementFromURI( ic->getUrl() );
         domCamera *c = daeSafeCast< domCamera >( el );
         if ( c == NULL )
         {
            osg::notify( osg::WARN ) << "Failed to locate camera " << ic->getUrl().getURI() << std::endl;
         }
         pGenericNode = processCamera( c );
         if ( pGenericNode != NULL )
         {
            pNewParent->asGroup()->addChild( pGenericNode );
         }
         continue;
      }

      domInstance_light *il = daeSafeCast< domInstance_light >( node->getContents()[i] );
      if ( il != NULL )
      {
         daeElement *el = getElementFromURI( il->getUrl() );
         domLight *l = daeSafeCast< domLight >( el );
         if ( l == NULL )
         {
            osg::notify( osg::WARN ) << "Failed to locate light " << il->getUrl().getURI() << std::endl;
         }
         pGenericNode = processLight( l );
         if ( pGenericNode != NULL )
         {
            pNewParent->asGroup()->addChild( pGenericNode );
         }
         continue;
      }


      // AMD Ignore the transformations.
      domTranslate *pTranslate = daeSafeCast< domTranslate >( node->getContents()[i] );
      if ( pTranslate != NULL )
      {
         continue;
      }

      domRotate *pRot = daeSafeCast< domRotate >( node->getContents()[i] );
      if ( pRot != NULL ) 
      {
         continue;
      }

      domScale *pScale = daeSafeCast< domScale >( node->getContents()[i] );
      if ( pScale != NULL ) 
      {
         continue;
      }

      domMatrix *pMatrix = daeSafeCast< domMatrix >( node->getContents()[i] );
      if ( pMatrix != NULL ) 
      {
         continue;
      }

      domSkew *pSkew = daeSafeCast< domSkew >( node->getContents()[i] );
      if ( pSkew != NULL ) 
      {
         continue;
      }

      // Catch any nodes we do not recognize.
      const char *name = node->getContents()[i]->getElementName();
      if ( name == NULL )
      {
         name = node->getContents()[i]->getTypeName();
      }
      osg::notify( osg::WARN ) << "Unsupported element type: " << name << " in COLLADA scene!" << std::endl;
   }
}
