/*
* Copyright 2006 Sony Computer Entertainment Inc.
*
* Licensed under the SCEA Shared Source License, Version 1.0 (the "License"); you may not use this 
* file except in compliance with the License. You may obtain a copy of the License at:
* http://research.scea.com/scea_shared_source_license.html
*
* Unless required by applicable law or agreed to in writing, software distributed under the License 
* is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or 
* implied. See the License for the specific language governing permissions and limitations under the 
* License. 
*/

#ifndef _DAE_CONV_H_
#define _DAE_CONV_H_

#include <string>

#include <dae.h>
#include <dae/daeURI.h>
#include <dae/daeElement.h>
#include <dom/domCommon_color_or_texture_type.h>

// AMD added
#include <dae/domAny.h>
#include <dom/domCOLLADA.h>
#include <dom/domProfile_COMMON.h>
#include <dom/domProfile_GLSL.h>
#include <osg/BlendFunc>
#include <dom/domInputLocalOffset.h>
// End AMD

#include <osg/Node>
#include <osg/Transform>
#include <osg/Notify>
#include <osg/PositionAttitudeTransform>
#include <osgDB/ReaderWriter>
#include <osgDB/FileNameUtils>
#include <osgDB/FileUtils>
#include <osgDB/Registry>
#include <osg/Material>
#include <osg/Texture>

// AMD added
#include <osg/Texture2D>

#define TANGENT_TEX_UNIT 6  //Tex texture unit to hold Tangent-Coords
#define BITANGENT_TEX_UNIT 7  //The texture unit to hold Bitangent-Coords. Sometimes called Binormal.
#define NM_TEX_UNIT 0 //The texture unit which holds UV for NORMALMAP in TangentSpace-Generation

class ColladaEffectNode;
class ColladaRTT;
// End AMD

class domBind_material;
class domCamera;
//class domCommon_color_or_texture_type; // AMD commented out
class domCommon_float_or_param_type;
class domGeometry;
class domInstance_controller;
class domInstance_geometry;
class domInstanceWithExtra;
class domLight;
class domLookat;
class domMatrix;
class domNode;
class domP;
class domProfile_COMMON;
class domScale;
class domSkew;
class domTranslate;
class domRotate;
class domVisual_scene;

// AMD commented out
//HF
//#include <dom/domProfile_GLSL.h>
//#include <osg/BlendFunc>
//#include <dom/domInputLocalOffset.h>
// End AMD


namespace osgdae 
{
   class domSourceReader;

   inline daeElement *getElementFromURI( daeURI &uri )
   {
      if ( uri.getState() == daeURI::uri_loaded || uri.getState() == daeURI::uri_pending ) {
         uri.resolveElement();
      }
      return uri.getElement();
   }
   inline daeElement *getElementFromIDRef( daeIDRef &idref )
   {
      if ( idref.getState() == daeIDRef::id_loaded || idref.getState() == daeIDRef::id_pending ) {
         idref.resolveElement();
      }
      return idref.getElement();
   }

   template< typename TInputArray, typename TInputType >
   bool findInputSourceBySemantic( TInputArray& inputs, const char* semantic, daeElement *& element, 
      TInputType ** input = NULL, int unit = 0 )
   {
      element = NULL;
      int count = 0;
      for ( size_t i = 0; i < inputs.getCount(); i++ ) {
         if ( !strcmp(semantic, inputs[i]->getSemantic()) ) {
            if ( count == unit )
            {
               element = getElementFromURI( inputs[i]->getSource() );
               *input = (TInputType*)inputs[i];
               return true;
            }
            count++;
         }
      }
      return false;
   }

   //HF
   inline const std::string getFilenameFromURI(daeURI& uri) 
   {
      uri.validate();

      char* path = NULL;
      char* filename = NULL;
      if ( std::string( uri.getProtocol() ) == std::string( "file" ) )
      {
         size_t bufSize = 1; //for the null char
         if ( uri.getFilepath() != NULL )
         {
            bufSize += strlen( uri.getFilepath() );
         }
         if ( uri.getFile() != NULL )
         {
            bufSize += strlen( uri.getFile() );
         }
         path = new char[bufSize+1];
         if ( !uri.getPath( path, daeInt(bufSize) ) )
         {
            osg::notify( osg::WARN ) << "Unable to get Path from URI" << std::endl;
            return "";
         }

#ifdef WIN32
         // under windows the URI passed back from the dom includes a / i.e. /c:/path/file, so we need to jump over the leading /
         filename = path+1;
#else
         filename = path;
#endif
      } 
      else 
      {
         osg::notify( osg::WARN ) << "Wrong Protocol for File-URI" << std::endl;
         return "";
      }

      std::string filenameStr( filename );

      if ( path != NULL ) 
      {
         delete[] path;	
      }

      return filenameStr;
   }

   /**
   @class daeReader
   @brief Read a OSG scene from a DAE file 
   */ 
   class daeReader 
   {

   public:

      daeReader( DAE *dae_ );

      virtual ~daeReader();

      bool convert( const std::string &fileURI );

      osg::Node* getRootNode()    
      { 
         return m_refRootNode.get();
      }

   protected:
      //scene processing
      osg::Node* processVisualScene( domVisual_scene *scene );

      void processNode( domNode *node, osg::Node *pParentNode );

      //transform processing
      osg::Transform* processMatrix( domMatrix *mat );
      osg::Transform* processTranslate( domTranslate *trans );
      osg::Transform* processRotate( domRotate *rot );
      osg::Transform* processScale( domScale *scale );
      osg::Transform* processLookat( domLookat *la );
      osg::Transform* processSkew( domSkew *skew );

      //geometry processing
      void processInstance_geometry( domInstance_geometry *ig, osg::Node *pParent ); //AMD change

      osg::Node* processGeometry( domGeometry *geo, ColladaEffectNode *pParent ); // AMD change
      
      void processInstance_controller( domInstance_controller *ictrl, osg::Node *pParent ); //AMD change

      typedef std::map< daeElement*, domSourceReader > SourceMap;
      typedef std::map< int, osg::IntArray*, std::less<int> > IndexMap;

      template< typename T >
      osg::Node* processSinglePPrimitive( T *group, SourceMap &sources, GLenum mode );

      template< typename T >
      osg::Node* processMultiPPrimitive( T *group, SourceMap &sources, GLenum mode );

      osg::Node* processPolylist( domPolylist *group, SourceMap &sources );

      void resolveArrays( domInputLocalOffset_Array &inputs, osg::Geometry *&geom, 
         SourceMap &sources, IndexMap &index_map );

      void processP( domP *p, osg::Geometry *&geom, IndexMap &index_map, osg::DrawArrayLengths* dal/*GLenum mode*/ );

      // AMD Change
      ColladaEffectNode * processBindMaterialUsingFX( domBind_material *bm, osg::Node *pParent );

      // AMD add
      ColladaEffectNode  * processMaterialUsingFX(  domMaterial             *mat, 
                                                domInstance_materialRef matRef,
                                                osg::Node               *pParent );

      ColladaEffectNode   *processEffectUsingFX(   domEffect                  *pDomEffect, 
                                                   domInstance_materialRef    rDomMaterialRef,
                                                   domMaterial                *pDomMaterial,
                                                   osg::Node                  *pParent);


      bool OverRideUniforms ( osg::StateSet* ss, domMaterial *mat, domEffect *effect );

      bool OverRideUniforms ( osg::StateSet* ss, domMaterial *mat, domProfile_GLSL::domTechnique::domPass* );
      // End AMD add

      osg::StateSet          *processProfileCOMMON( domProfile_COMMON *pc );      

      /************************************************************************/
      /* Additional Method for handling profileGLSL  - HF                     */
      /************************************************************************/

      // AMD change
      ColladaEffectNode * processProfileGLSLUsingFX(  domProfile_GLSL            *pGlsl, 
                                                      domInstance_materialRef    rDomMaterialRef, 
                                                      domMaterial                *pDomMaterial,
                                                      domEffect                  *pDomEffect,
                                                      osg::Node                  *pParent);
      bool processBindArray(  osg::StateSet  *, 
                              domProfile_GLSL::domTechnique::domPass::domShader  * shaderEl, 
                              domProfile_GLSL::domTechnique::domPass *  pDomPass,
                              domMaterial                            *  pDomMaterial,
                              //bool                                      updateSemanticBindsOnly = false,
                              ColladaEffectNode                      *  pEffect = NULL );

      bool GetStateSet( domProfile_GLSL::domTechnique::domPass *pDomPass,
                        osg::StateSet                          *& out_pStateSet,
                        osg::Program                           *& out_pProgram );

      bool CheckEffectCache( domProfile_GLSL *pGlsl );

      osg::Shader* LoadShader ( daeElement *pResolvedElement, domGlsl_pipeline_stage eStage );

      daeElement* ResolveElement (  std::string &strSource,
         daeElement *pParent );

      void  AddVertexAttributes ( 	osg::Program   *sProgram, 
         domInstance_material::domBind_vertex_input_Array & rBindVertexInputArray );
      // End AMD change

      bool processGlPipelineSettings( osg::StateSet*, domGl_pipeline_settings* );

      template <typename T>
      bool processCommonParamTypes(T*, const char*, osg::Uniform*&);

      bool processGlslParamType(domGlsl_param_type*, const char*, osg::Uniform*& );
      bool processFxCommonParamType(domFx_basic_type_common*, const char*, osg::Uniform*& );

      //NOTE: this method also tries to upcast to glslSurfaceType

      //Resolves the surface element, creates the necessary osg::Image data, 
      //returns a osg::Texture...
      osg::Texture* getTextureFromSurface(const std::string& surfRef, daeElement* samplerEl);
      template <typename T>
      osg::Texture* processSurfaceTypes( T*, std::string strParamSID ); // AMD change

      //Applying texture-states from sampler-elements
      //Commond attributes for all sampler-types except DEPTH
      template<typename T>
      bool processSamplerCommon1D(osg::Texture*, T*);

      //sampler2D and samplerRect
      template<typename T>
      bool processSampler2D(osg::Texture*, T*);

      //sampler3D, samplerCUBE
      template<typename T>
      bool processSampler3D(osg::Texture*, T*);

      template<typename T>
      bool processSamplerDEPTH(osg::Texture*, T*);


      bool processColorOrTextureType(    domCommon_color_or_texture_type *cot, 
         osg::Material::ColorMode channel, 
         osg::Material *mat, 
         domCommon_float_or_param_type *fop = NULL, 
         osg::StateAttribute **sa = NULL );

      osg::StateAttribute *processTransparentType( domCommon_transparent_type *ctt, osg::StateSet *ss );

      osg::StateAttribute *processTexture( domCommon_color_or_texture_type_complexType::domTexture *tex );

      //scene objects
      osg::Node* processLight( domLight *dlight );

      osg::Node* processCamera( domCamera *dcamera );

      // AMD change
      domImage* preCheckDOMImage ( daeIDRef &ref );

      bool CheckEvaluateScene( domMaterial  *pDomMaterial, ColladaRTT *pRTT );

      osg::Texture * GetRttTexture ( std::string strParamSID ); 

      template <typename T>
      bool  FindInExtra(   T           *  pElement, 
                           const char  *  pProfileName,
                           const char  *  pElementName,
                           const char  *  pParamID,
                           std::string &  rOutString );

      void HandleEffectExtras ( domEffect * pDomEffect, ColladaEffectNode * pEffect );

      bool OveriddenByMaterial ( const std::string &  rStrParamName,
                                 domMaterial       *  pDomMaterial,
                                 unsigned int      &  nParamIndex );

      domProfile_GLSL * GetGLSLProfile ( const domEffect *pCurrentEffect ) const;

      domGlsl_surface_type * getSurfaceBySid ( const std::string& sid ) const;

   
   protected: // DATA

      std::map< domGeometry*, osg::Node* >         m_geometryMap;

      // NOTE: The material map does not seem to be used for anything.
      std::map< domMaterial*, osg::StateSet* >     m_materialMap;
      // End AMD change

      //HF: map for node-material binding
      std::map< osg::Node*, std::string >          m_bindMaterialMap;
      std::map< domMaterial*, ColladaEffectNode* > m_materialEffectMap;

      //HF
      typedef std::map<std::string, std::string >                                ParamToUniformMap;
      typedef std::pair<osg::StateSet*, ParamToUniformMap>                       EffectData;
      typedef std::map<domProfile_GLSL*, osg::StateSet* >                        ProfileGLSLMap;
      // AMD
      // Performs same function as ProfileGLSLMap but works with Effects
      typedef std::map<domProfile_GLSL::domTechnique::domPass*, osg::StateSet* > PassGLSLMap;
      typedef std::map< std::string,  osg::Texture * >                           RTTTextureMap;

      typedef std::map<domImage*, osg::Image* >                                  ImageMap;
      

      //Maps from COLLADA DOM States to osg States
      std::map< domFx_sampler_wrap_common, osg::Texture::WrapMode >     m_wrapModeMap;
      std::map< domFx_sampler_filter_common, osg::Texture::FilterMode > m_filterModeMap;
      std::map< domGl_blend_type, osg::BlendFunc::BlendFuncMode >       m_blendFuncModeMap;
      // AMD 
      std::map< domEffect*, EffectData>                                 m_loadedEffects;
      std::map< domProfile_GLSL::domTechnique::domPass*, EffectData>    m_loadedPasses; 
      std::map< daeElement*, osg::Texture * >                           m_loadedSurfacesMap;
      // End AMD

      // Only load a domImage once
      ImageMap       m_ImageMap;
      RTTTextureMap  m_RTTTextureMap; // AMD

      // Profile_glsl cache
      ProfileGLSLMap m_profileGLSLCache;
      
      // AMD Add
      PassGLSLMap    m_passGLSLMap;
      DAE         *  m_pDae;
      domCOLLADA  *  m_pDoc ;
      osg::Node   *  m_pRootNode;
      int            m_numlights;
      domEffect   *  m_pCurrentEffect;

      osg::ref_ptr<osg::Node> m_refRootNode;

      ColladaEffectNode *m_pCurrentColladaEffect;

      // Pointers to current material. Used when binding params
      // to the shaders. They allow the loader code to search the setparam arrays in the technique 
      // and material to see if they are overriden. 
      domMaterial                   *  m_pCurrentDomMaterial;
      // End AMD Add

   };
};


// AMD add
template <typename T>
bool osgdae::daeReader::FindInExtra(   T           *  pElement, 
                                       const char  *  pProfileName,
                                       const char  *  pElementName,
                                       const char  *  pParamID,
                                       std::string &  rOutString )
{
   assert ( pElement ) ;
   assert ( pProfileName ) ;
   assert ( pElementName ) ;

   // For each extra
   int nExtraCount = pElement->getExtra_array().getCount();
   for ( int nExtraIndex = 0 ; nExtraIndex < nExtraCount; nExtraIndex++ ) 
   {
      // Get the extra.
      domExtra *extra = pElement->getExtra_array().get( nExtraIndex );
      if ( !extra )
      {
         return false ;
      }

      std::string strSearchProfile = pProfileName ;

      // For each technique
      int nTechCount = extra->getTechnique_array().getCount();
      for ( int nTechIndex = 0 ; nTechIndex < nTechCount; nTechIndex++ )
      {
         // Get the technique.
         domTechnique *pTech = extra->getTechnique_array().get( nTechIndex );
         if ( !pTech )
         {
            continue;
         }
         // Check to see if we have the RM profile.
         std::string strProfile = pTech->getProfile();
         if ( strProfile != strSearchProfile )
         {
            continue;
         }

         std::string strSearchElementName = pElementName;

         // Search each element in the contents.
         int nContentsCount = pTech->getContents().getCount();
         for ( int nContentsIndex = 0 ; nContentsIndex < nContentsCount ; nContentsIndex++ )
         {
            // Get the element
            daeElement *pElement = pTech->getContents().get( nContentsIndex );
            if ( !pElement ) 
            {
               continue;
            }
            // Check if this is the element we are looking for.
            std::string strElementName = pElement->getElementName();
            if ( strSearchElementName != strElementName )
            {
               continue;
            }
            // Get the children of the element
            daeTArray<daeSmartRef<daeElement> > pChildren; 
            pElement->getChildren( pChildren );
            int nChildCount = pChildren.getCount();
            if ( 0 == nChildCount )
            {
               continue;
            }
            // Look in each of the children.
            for ( int nChildIndex = 0 ;  nChildIndex < nChildCount ; nChildIndex++ )
            {
               domParam *pParam = daeSafeCast<domParam>(pChildren.get( nChildIndex ));
               if ( !pParam )
               {
                  continue;
               }
               
               // Search for a specific SID only if on is supplied. 
               if ( strlen ( pParamID ) )
               {
                  std::string strID = pParam->getSid();
                  std::string srcSearchID = pParamID;
                  if ( strID != srcSearchID ) 
                  {
                     continue;
                  }
               }

               // Get the string value.
               char ** pVal = (char **)pParam->getValuePointer();
               if ( pVal && *pVal )
               {
                  rOutString = *pVal;
                  return true;
               }
            }
         }
      }
   }
   return false;
}
// End AMD add

#endif


